/*
 * Java
 *
 * Copyright 2008-2022 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.elflw.input;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;

import com.is2t.elflw.ElfConstants;
import com.is2t.elflw.ElfGenerator;
import com.is2t.elflw.allocator.MemoryBlock;
import com.is2t.elflw.nodes.AllocationSection;
import com.is2t.elflw.nodes.DynamicAllocationSection;
import com.is2t.elflw.nodes.ElfRelocatableUnit;
import com.is2t.elflw.nodes.ElfUnit;
import com.is2t.elflw.nodes.ProgAllocationSection;
import com.is2t.elflw.nodes.ProgControlSection;
import com.is2t.elflw.nodes.ProgSection;
import com.is2t.elflw.nodes.RelocationEntry;
import com.is2t.elflw.nodes.RelocationSection;
import com.is2t.elflw.nodes.Section;
import com.is2t.elflw.nodes.StringSection;
import com.is2t.elflw.nodes.SymbolTableEntry;
import com.is2t.elflw.nodes.SymbolTableSection;
import com.is2t.elflw.nodes.TargetSpecificSection;
import com.is2t.elflw.support.Header;
import com.is2t.elflw.support.SectionHeader;

import ej.annotation.Nullable;

/**
 * Implementation of Executable and Linking Format (ELF) Specification. See Sun Libraries and Linker Guide, February
 * 2007 (819-0960.pdf), Chapter 9 : Object File Format
 *
 * This library is a lightweight implementation based on the original IS2T ELF desktop library used in linker. It only
 * loads ELF relocatable files that have standard sections (no machine specific handling).
 * <ul>
 * <li>No support for Group sections</li>
 * <li>32 bits oriented</li>
 * <li>Load from an {@link InputStream}</li>
 * <li>Char arrays are not intern (no symbols table)</li>
 * <li>Control sections are created but content and relocations are skipped</li>
 * <li>Section names are skipped</li> ELF constraints:
 * <li>Assumes that the sections header appears right after the header. (i.e. the ELF binary file may have been
 * generated using IS2T linker with generateElfForStreaming option set to true</li>
 * <li>ELF file has One and only one {@link SymbolTableSection}</li>
 * </ul>
 */
public abstract class AbstractElfLoader
		implements ElfConstants, InconsistentLibraryErrorConstants, ErrorMessagesConstants, ElfGenerator {

	@Nullable
	public ElfRelocatableUnit currentUnit;
	public boolean lsbAndNotMsb;

	// section headers
	@Nullable
	protected SectionHeader[] sectionHeaders;
	// current index of the header for the section being loaded
	protected int currentSectionIndex;

	public AbstractElfLoader() {
	}

	/**
	 * return the code of the supported machine for this loader
	 */
	public abstract int supportedMachine();

	/**
	 * Gets a tag name that uniquely identifies this loader.
	 *
	 * @return an intern {@link String}
	 */
	public abstract String getTag();

	/**
	 * Tell whether this seems to be an ELF file.
	 *
	 * @param bytes
	 *            the first 4 bytes of the stream
	 * @return true if this is an ELF file, false otherwise.
	 */
	public static boolean canLoad(byte[] bytes) {
		if (bytes.length < SIGNATURE_LENGTH) {
			return false;
		}

		// read magic fields
		if (!(bytes[EI_MAG0] == ELFMAG0 && bytes[EI_MAG1] == ELFMAG1 && bytes[EI_MAG2] == ELFMAG2
				&& bytes[EI_MAG3] == ELFMAG3)) {
			// not ELF signature
			return false;
		}

		return true;
	}

	private int loadHeader(SectionHeader h, byte[] bytes, int ptr) {
		h.name = u4(bytes, ptr);
		ptr += 4;
		h.type = u4(bytes, ptr);
		ptr += 4;
		h.flags = u4(bytes, ptr);
		ptr += 4;
		h.addr = u4(bytes, ptr);
		ptr += 4;
		h.offset = u4(bytes, ptr);
		ptr += 4;
		h.size = u4(bytes, ptr);
		ptr += 4;
		h.link = u4(bytes, ptr);
		ptr += 4;
		h.info = u4(bytes, ptr);
		ptr += 4;
		h.addralign = u4(bytes, ptr);
		ptr += 4;
		h.entsize = u4(bytes, ptr);
		ptr += 4;
		return ptr;
	}

	/**
	 * Load a Relocatable unit. Assume the ELF magic field has been read (the 4 bytes).
	 */
	public ElfRelocatableUnit load(byte[] elfSignature, InputStream is) throws IOException, ElfLoaderError {
		// check again the signature
		if (!canLoad(elfSignature)) {
			throw new ElfLoaderError().invalidELFSignature(elfSignature);
		}

		Header h = loadHeader(is);

		// Currently, the ELF loader is only designed for ELF32 files
		if (h.e_ident_class != ELFCLASS32) {
			throw new ElfLoaderError().inconsistentLibrary(ELF64CLASS);
		}

		// load sections headers

		// Constraint: this loader assumes that section header appears right after the program header
		// which is not mandatory in ELF specification (see class comment)
		if (h.e_shoff != Sizeof32bitsHeader) {
			throw new ElfLoaderError().inconsistentLibrary(ExpectedSectionsHeadersRightAfterUnitHeader);
		}

		// At least one section header at index 0 that represent no section
		int absoluteBytesPtr = Sizeof32bitsHeader;
		int size = Sizeof32bitsSectionHeader;
		byte[] bytes = new byte[size];
		readFully(is, bytes, 0, bytes.length);
		absoluteBytesPtr += size;
		int ptr = 0;
		SectionHeader firstSection = new SectionHeader();
		ptr = loadHeader(firstSection, bytes, ptr);

		int nbSections = h.e_shnum;
		if (nbSections == SHN_UNDEF) {
			// >= SHN_LORESERVE sections : nbSections is in the sh_size field of the first section
			nbSections = firstSection.size;
		}

		int strTableHeaderIndex = h.e_shstrndx;
		if (strTableHeaderIndex == SHN_XINDEX) {
			strTableHeaderIndex = firstSection.link;
		}

		size = Sizeof32bitsSectionHeader * (nbSections - 1);
		bytes = new byte[size];
		readFully(is, bytes, 0, bytes.length);
		absoluteBytesPtr += size;
		ptr = 0;

		SectionHeader strTableHeader = null;
		SectionHeader[] sectionHeaders = new SectionHeader[nbSections];
		sectionHeaders[0] = firstSection;
		for (int i = 0; ++i < nbSections;) {
			SectionHeader sh = sectionHeaders[i] = new SectionHeader();
			if (i == strTableHeaderIndex) {
				strTableHeader = sh;
			}
			ptr = loadHeader(sh, bytes, ptr);
		}

		if (strTableHeader == null) {
			// no string section for section names !
			// Seems to be an error is there is more that 1 section (how we can get the names ??)
			if (sectionHeaders.length > 1) {
				throw new ElfLoaderError().inconsistentLibrary(NoStringTableForHeaderButMoreThanOneSection);
			}
		}

		// Here, all section headers have been loaded
		sectionHeadersLoaded(sectionHeaders);

		// Create the unit
		ElfRelocatableUnit lu = this.currentUnit = new ElfRelocatableUnit(h.e_machine, h.lsbAndNotMsb(),
				h.e_ident_abiversion, h.e_flags);

		// Load sections [1..NbSections-1] (only allocation sections)
		// the array may contain null entries
		Section[] sections = new Section[nbSections - 1];
		int[] absoluteBytesPtrRef = new int[] { absoluteBytesPtr };
		for (int i = 0; ++i < nbSections;) { // skip first section header
			SectionHeader sh = sectionHeaders[i];
			assert sh != null;
			Section section = createSection(i, sh, absoluteBytesPtrRef, is);
			sections[i - 1] = section;
		}

		lu.sections = sections;

		// All sections have been created => load sections
		this.sectionHeaders = sectionHeaders;
		for (int i = sections.length; --i >= 0;) {
			this.currentSectionIndex = i + 1;
			sections[i].generateUsing(this);
		}

		if (lu.symbolTableSection == null) {
			throw new ElfLoaderError().inconsistentLibrary(ExpectedASymbolTableSection);
		}

		// Here, the ELF unit is successfully loaded
		return lu;
	}

	/**
	 * Callback when section headers are loaded, before loading the sections content
	 *
	 * @param the
	 *            raw ELF section headers, including the "null" section at index 0
	 */
	protected void sectionHeadersLoaded(SectionHeader[] sectionHeaders) {
		// NOP
	}

	protected SectionHeader currentSectionHeader() {
		SectionHeader[] sectionHeaders = this.sectionHeaders;
		assert sectionHeaders != null;
		int currentSectionIndex = this.currentSectionIndex;
		assert currentSectionIndex > 0;
		SectionHeader sh = sectionHeaders[currentSectionIndex];
		assert sh != null;
		return sh;
	}

	private void generateSaveAndRestoreContext(Section section, int index) {
		// Should be called when a section is generated recursively
		int previous = this.currentSectionIndex;
		this.currentSectionIndex = index;
		section.generateUsing(this);
		this.currentSectionIndex = previous;
	}

	public byte[] extractDataAsByteArray(SectionHeader sh, int[] absolutePtrRef, InputStream is) throws IOException {
		// Extract the data of the section described by the given section header
		assert sh.type != SHT_NOBITS;
		byte[] data = new byte[sh.size];
		readFully(is, data, 0, data.length);
		absolutePtrRef[0] += data.length;
		return data;
	}

	protected void transferSectionDataToMemoryBlock(ProgAllocationSection section, int[] absolutePtrRef, InputStream is)
			throws IOException {
		// Extract the data of the section described by the given section header
		byte[] bytes = new byte[getTransferBufferSize()];
		int size = section.memSize();
		MemoryBlock block = section.getMemoryBlock();
		int totalNbRead = 0;
		while (totalNbRead < size) {
			int nbRead = is.read(bytes, 0, Math.min(size - totalNbRead, bytes.length));
			if (nbRead == -1) {
				throw new EOFException();
			}
			boolean copyOK = block.copyFrom(bytes, 0, totalNbRead, nbRead);
			if (!copyOK) {
				onCopyError();
			}
			totalNbRead += nbRead;
		}
		absolutePtrRef[0] += size;
	}

	/**
	 * Gets the size in bytes of the temporary buffer allocated in heap used to transfer bytes directly from ELF
	 * {@link InputStream} to target {@link MemoryBlock}
	 *
	 * @return the size in bytes
	 */
	protected int getTransferBufferSize() {
		return 512;
	}

	/**
	 * Create and load a section
	 *
	 * @param absolutePtrRef
	 *            is updated Subclasses may override it
	 */
	public Section createSection(int sectionIndex, SectionHeader sh, int[] absolutePtrRef, InputStream is)
			throws IOException {
		// skip padding bytes
		int padding = sh.offset - absolutePtrRef[0];
		if (padding < 0) {
			throw new ElfLoaderError().inconsistentLibrary(NegativePadding);
		} else if (padding > 0) {
			readFully(is, new byte[padding], 0, padding);
		}
		absolutePtrRef[0] += padding;

		int flags = sh.flags;
		switch (sh.type) {
		case SHT_SYMTAB: {
			Section s = new SymbolTableSection(sh.addralign, extractDataAsByteArray(sh, absolutePtrRef, is));
			return s;
		}
		case SHT_PROGBITS: {
			if ((flags & SHF_ALLOC) == SHF_ALLOC) {
				ProgAllocationSection section = newProgAllocationSection(sectionIndex, sh);
				transferSectionDataToMemoryBlock(section, absolutePtrRef, is);
				return section;
			} else {
				// control section: data not loaded
				return new ProgControlSection(sh.addralign, sh.size);
			}
		}
		case SHT_STRTAB: {
			return new StringSection(sh.addralign, extractDataAsByteArray(sh, absolutePtrRef, is));
		}
		case SHT_REL:
		case SHT_RELA: {
			Section s = new RelocationSection(sh.addralign, sh.type == SHT_RELA,
					extractDataAsByteArray(sh, absolutePtrRef, is));
			return s;
		}
		case SHT_NOBITS: {
			if ((flags & SHF_ALLOC) == SHF_ALLOC) {
				DynamicAllocationSection section = newDynamicAllocationSection(sectionIndex, sh);
				return section;
			} else {
				throw new ElfLoaderError().unknownSectionType(sectionIndex, sh.type);
			}
		}
		case SHT_NOTE:
		case SHT_NULL:
		case SHT_HASH:
		case SHT_DYNAMIC:
		case SHT_SHLIB:
		case SHT_DYNSYM:
		case SHT_INIT_ARRAY:
		case SHT_FINI_ARRAY:
		case SHT_PREINIT_ARRAY:
		case SHT_SYMTAB_SHNDX:
		case SHT_GROUP:
		default:
			byte[] bytes = extractDataAsByteArray(sh, absolutePtrRef, is);
			Section s = createCustomSection(sh, bytes);
			if (s == null) {
				throw new ElfLoaderError().unknownSectionType(sectionIndex, sh.type);
			} else {
				return s;
			}
		}
	}

	/**
	 * Creates a {@link ProgAllocationSection} with its allocated {@link MemoryBlock}.
	 *
	 * @param sectionIndex
	 *            the ELF section index
	 * @param sectionHeader
	 *            the ELF section header
	 * @return a new {@link ProgAllocationSection}
	 */
	protected abstract ProgAllocationSection newProgAllocationSection(int sectionIndex, SectionHeader sectionHeader);

	/**
	 * Creates a {@link DynamicAllocationSection} with allocated {@link MemoryBlock}.
	 *
	 * @param sectionIndex
	 *            the ELF section index
	 * @param sectionHeader
	 *            the ELF section header
	 * @return a new {@link DynamicAllocationSection}
	 */
	protected abstract DynamicAllocationSection newDynamicAllocationSection(int sectionIndex,
			SectionHeader sectionHeader);

	/**
	 * This method is called when {@link MemoryBlock#copyFrom(byte[], int, int, int)} has returned <code>false</code>
	 */
	protected abstract void onCopyError();

	/**
	 * Creates a custom section.
	 *
	 * @param sh
	 *            the section header
	 * @param bytes
	 *            the section bytes
	 * @return the created section or <code>null</code> if not handled
	 */
	@Nullable
	protected Section createCustomSection(SectionHeader sh, byte[] bytes) {
		return null;
	}

	/**
	 * Get the non null section at the given index.
	 *
	 * @return the non null section at the given index
	 */
	public Section getExpectedSection(int index) {
		ElfRelocatableUnit currentUnit = this.currentUnit;
		assert currentUnit != null;
		Section s = currentUnit.section(index);
		assert s != null;
		return s;
	}

	@Override
	public void generateRelocationSection(RelocationSection rs) {
		SectionHeader sh = currentSectionHeader();

		// Load link and info header fields (See Table7-9)
		int targetSectionIndex = sh.info;
		ProgSection targetSection = rs.targetSection = getExpectedProgSection(targetSectionIndex);

		SymbolTableSection sbs = rs.symbolTable = (SymbolTableSection) getExpectedSection(sh.link);
		generateSaveAndRestoreContext(sbs, sh.link); // ensure the symbol table is loaded

		if (!(targetSection instanceof AllocationSection)) {
			// skipped
			return;
		}
		if ((rs.withAddend && sh.entsize != SizeofRelocationEntryAddend)
				|| (!rs.withAddend && sh.entsize != SizeofRelocationEntry)) {
			throw new ElfLoaderError().inconsistentLibrary(WrongExpectedEntrySize);
		}
	}

	protected ProgSection getExpectedProgSection(int targetSectionIndex) {
		try {
			return (ProgSection) getExpectedSection(targetSectionIndex);
		} catch (ClassCastException e) {
			throw new ElfLoaderError().inconsistentLibrary(NotAProgBitsSection);
		}
	}

	/**
	 * Load the ELF header starting right after the ELF magic field (Assume the first 4 bytes has been read).
	 *
	 * @throws ElfLoaderError
	 *             if the header is invalid
	 */
	public Header loadHeader(InputStream is) throws IOException {
		Header header = new Header();

		byte[] bytes = new byte[Sizeof32bitsHeader];
		readFully(is, bytes, SIGNATURE_LENGTH, bytes.length - SIGNATURE_LENGTH);

		header.e_ident_class = bytes[EI_CLASS];
		if (header.e_ident_class != ELFCLASS32 && header.e_ident_class != ELFCLASS64) {
			throw new ElfLoaderError().invalidHeader(InternalWrong_ident_class);
		}

		header.e_ident_data = bytes[EI_DATA];
		if (header.e_ident_data != ELFDATA2LSB && header.e_ident_data != ELFDATA2MSB) {
			throw new ElfLoaderError().invalidHeader(InternalWrong_ident_data);
		}
		this.lsbAndNotMsb = (header.e_ident_data == ELFDATA2LSB);

		header.e_ident_version = bytes[EI_VERSION];
		if (header.e_ident_version != EV_CURRENT) {
			throw new ElfLoaderError().invalidHeader(InternalWrong_ident_version);
		}

		header.e_ident_osabi = bytes[EI_OSABI];

		header.e_ident_abiversion = bytes[EI_ABIVERSION];

		int ptr = EI_NIDENT;
		header.e_type = u2(bytes, ptr);
		ptr += 2;
		if (header.e_type != ET_REL) {
			throw new ElfLoaderError().invalidHeaderType(header.e_type, ET_REL);
		}

		int value;
		header.e_machine = value = u2(bytes, ptr);
		ptr += 2;
		if (header.e_machine != supportedMachine()) {
			throw new ElfLoaderError().unsupportedArchitecture(header.e_machine, supportedMachine());
		}

		header.e_version = value = u4(bytes, ptr);
		ptr += 4;
		if (value == EV_NONE) {
			throw new ElfLoaderError().invalidHeader(InternalWrong_version);
		}
		header.e_entry = u4(bytes, ptr);
		ptr += 4;
		header.e_phoff = u4(bytes, ptr);
		ptr += 4;
		header.e_shoff = u4(bytes, ptr);
		ptr += 4;
		header.e_flags = u4(bytes, ptr);
		ptr += 4;
		header.e_ehsize = u2(bytes, ptr);
		ptr += 2;
		header.e_phentsize = u2(bytes, ptr);
		ptr += 2;
		header.e_phnum = u2(bytes, ptr);
		ptr += 2;
		header.e_shentsize = u2(bytes, ptr);
		ptr += 2;
		header.e_shnum = u2(bytes, ptr);
		ptr += 2;
		header.e_shstrndx = u2(bytes, ptr);
		ptr += 2;

		return header;
	}

	private void readFully(InputStream is, byte[] bytes, int offset, int length) throws IOException {
		int end = offset + length;
		while (offset < end) {
			int nbRead = is.read(bytes, offset, end - offset);
			if (nbRead == -1) {
				throw new EOFException();
			}
			offset += nbRead;
		}
	}

	public static int u1(byte[] bytes, int ptr) {
		return bytes[ptr] & 0xFF;
	}

	public int u2(byte[] bytes, int ptr) {
		return u2(bytes, ptr, this.lsbAndNotMsb);
	}

	public int u4(byte[] bytes, int ptr) {
		return u4(bytes, ptr, this.lsbAndNotMsb);
	}

	public static int u2(byte[] bytes, int ptr, boolean lsbAndNotMsb) {
		if (lsbAndNotMsb) {
			return (bytes[ptr++] & 0xFF) | (((bytes[ptr++] & 0xFF)) << 8);
		} else {
			return (((bytes[ptr++] & 0xFF)) << 8) | (bytes[ptr++] & 0xFF);
		}
	}

	public static int u4(byte[] bytes, int ptr, boolean lsbAndNotMsb) {
		if (lsbAndNotMsb) {
			return (bytes[ptr++] & 0xFF) | (((bytes[ptr++] & 0xFF)) << 8) | (((bytes[ptr++] & 0xFF)) << 16)
					| (((bytes[ptr] & 0xFF)) << 24);
		} else {
			return (((bytes[ptr++] & 0xFF)) << 24) | (((bytes[ptr++] & 0xFF)) << 16) | (((bytes[ptr++] & 0xFF)) << 8)
					| (bytes[ptr] & 0xFF);
		}
	}

	public static void writeU4(byte[] bytes, int ptr, boolean lsbAndNotMsb, int u4) {
		if (lsbAndNotMsb) {
			bytes[ptr++] = (byte) u4;
			bytes[ptr++] = (byte) (u4 >>> 8);
			bytes[ptr++] = (byte) (u4 >>> 16);
			bytes[ptr++] = (byte) (u4 >>> 24);
		} else {
			bytes[ptr++] = (byte) (u4 >>> 24);
			bytes[ptr++] = (byte) (u4 >>> 16);
			bytes[ptr++] = (byte) (u4 >>> 8);
			bytes[ptr++] = (byte) u4;
		}
	}

	@Override
	public void generateSymbolTableSection(SymbolTableSection node) {
		if (node.stringSection != null) {
			// already generated
			return;
		}
		SectionHeader header = currentSectionHeader();

		StringSection stringTable = (StringSection) getExpectedSection(header.link);
		generateSaveAndRestoreContext(stringTable, header.link); // ensure stringTable is loaded
		node.stringSection = stringTable;

		if (header.entsize != SizeofSymbolTableEntry) {
			throw new ElfLoaderError().inconsistentLibrary(WrongExpectedEntrySize);
		}

		ElfRelocatableUnit currentUnit = this.currentUnit;
		assert currentUnit != null;
		if (currentUnit.symbolTableSection != null) {
			throw new ElfLoaderError().inconsistentLibrary(ExpectedASymbolTableSection);
		}
		currentUnit.symbolTableSection = node;
	}

	// "nothing to load" methods
	@Override
	public void generateStringSection(StringSection section) {
	}

	@Override
	public void generateDynamicAllocationSection(DynamicAllocationSection section) {
	}

	@Override
	public void generateProgAllocationSection(ProgAllocationSection node) {
	}

	@Override
	public void generateProgControlSection(ProgControlSection node) {
	}

	@Override
	public void generateTargetSpecificSection(TargetSpecificSection node) {
		/* default behavior */}

	// unused methods
	@Override
	public void generateElfRelocatableUnit(ElfRelocatableUnit unit) {
		throw new RuntimeException();
	}

	@Override
	public void generateRelocationEntry(RelocationEntry entry) {
		throw new RuntimeException();
	}

	@Override
	public void generateSymbolTableEntry(SymbolTableEntry node) {
		throw new RuntimeException();
	}

	/**
	 * Load {@link RelocationEntry} on the fly
	 *
	 * @param ptr
	 *            offset in {@link RelocationSection#data}
	 */
	public static RelocationEntry loadRelocationEntry(ElfUnit unit, RelocationSection section, int entryIndex) {
		int ptr = entryIndex * section.entrySize();
		boolean lsbAndNotMsb = unit.lsbAndNotMsb;
		byte[] bytes = section.data;
		int r_offset = u4(bytes, ptr, lsbAndNotMsb);
		ptr += 4;
		int r_info = u4(bytes, ptr, lsbAndNotMsb);
		ptr += 4;
		SymbolTableSection symbolTable = section.symbolTable;
		assert symbolTable != null;
		SymbolTableEntry symbol = loadSymbolTableEntry(unit, symbolTable, RelocationEntry.ELF32_R_SYM(r_info));

		int relocType = RelocationEntry.ELF32_R_TYPE(r_info);
		int addend;

		if (section.withAddend) {
			addend = u4(bytes, ptr, lsbAndNotMsb);
			ptr += 4;
		} else {
			addend = 0xCAFECAFE;
		}
		return new RelocationEntry(r_offset, symbol, relocType, addend);
	}

	/**
	 * @return <code>null</code> if symbolIndex is <code>0</code>
	 */
	@Nullable
	public static SymbolTableEntry loadSymbolTableEntry(ElfUnit unit, SymbolTableSection section, int symbolIndex) {
		if (symbolIndex == 0) {
			return null;
		}
		StringSection stringTable = section.stringSection;
		assert stringTable != null;
		boolean lsbAndNotMsb = unit.lsbAndNotMsb;
		byte[] bytes = section.data;
		int ptr = symbolIndex * SizeofSymbolTableEntry;
		int nameIndex = u4(bytes, ptr, lsbAndNotMsb);
		ptr += 4;
		byte[] name = stringTable.getName(nameIndex);
		int value = u4(bytes, ptr, lsbAndNotMsb);
		ptr += 4;
		int size = u4(bytes, ptr, lsbAndNotMsb);
		ptr += 4;
		int info = u1(bytes, ptr);
		++ptr;
		int other = u1(bytes, ptr);
		++ptr;
		int sectionIndex = u2(bytes, ptr, lsbAndNotMsb);
		ptr += 2;

		Section targetSection = unit.section(sectionIndex);
		if (targetSection == null) {
			return new SymbolTableEntry(name, value, size, info, other, sectionIndex);
		} else {
			return new SymbolTableEntry(name, value, size, info, other, targetSection);
		}
	}
}
