/*
 * Java
 *
 * Copyright 2008-2022 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.elflw.nodes;

import com.is2t.elflw.ElfConstants;
import com.is2t.elflw.ElfGenerator;

import ej.annotation.Nullable;

/**
 * Symbol entry in a symbol table section
 */
public class SymbolTableEntry extends ElfLinkerNode implements ElfConstants {

	private static final int INFO_BINDING_OFFSET = 4;
	private static final int INFO_BINDING_MASK = 0xF;
	private static final int INFO_TYPE_OFFSET = 0;
	private static final int INFO_TYPE_MASK = 0xF;

	public static final int NoMeanSectionIndex = -1;

	public byte[] name;

	// When targetSection != null, value is the offset in the section.
	// Otherwise it is the absolute value of the symbol
	public int value;

	// Size is the size of the value targeted by the symbol
	public int size;

	public int info; // see getBinding()/getType()
	public int other; // see getVisibility()

	/**
	 * Has a mean only when {@link #targetSection == null} Value = {@link ElfConstants#SHN_UNDEF} or between
	 * {@link ElfConstants#SHN_LORESERVE} and {@link ElfConstants#SHN_HIRESERVE}
	 */
	public int sectionIndex;

	/**
	 * When != null, this symbol entry refers to a section (sectionIndex has no mean)
	 */
	@Nullable
	public Section targetSection; // may be != null when sectionIndex != SNH_UNDEF & the entry refer to a section

	public SymbolTableEntry(byte[] name, int value, int size, int info, int other, Section targetSection) {
		this(name, value, size, info, other, NoMeanSectionIndex, targetSection);
	}

	public SymbolTableEntry(byte[] name, int value, int size, int info, int other, int sectionIndex) {
		this(name, value, size, info, other, sectionIndex, null);
	}

	private SymbolTableEntry(byte[] name, int value, int size, int info, int other, int sectionIndex,
			@Nullable Section targetSection) {
		this.name = name;
		this.value = value;
		this.size = size;
		this.info = info;
		this.other = other;
		this.sectionIndex = sectionIndex;
		this.targetSection = targetSection;
	}

	/**
	 * Return STT_XXX type
	 */
	public int getType() {
		return (this.info >>> INFO_TYPE_OFFSET) & INFO_TYPE_MASK;
	}

	/**
	 * Return STB_XXX binding
	 */
	public int getBinding() {
		return (this.info >>> INFO_BINDING_OFFSET) & INFO_BINDING_MASK;
	}

	/**
	 * Return STV_XXX visibility
	 */
	public int getVisibility() {
		return this.other & 0x3;
	}

	/**
	 * Set symbol STV_XXX visibility
	 */
	public void setVisibility(int visibility) {
		this.other = ((this.other & ~0x3) | visibility);
	}

	public void setSize(int size) {
		this.size = size;
	}

	public void setType(int type) {
		this.info = (this.info & ~(INFO_TYPE_MASK << INFO_TYPE_OFFSET)) | (type << INFO_TYPE_OFFSET);
	}

	private void setBinding(int binding) {
		this.info = (this.info & ~(INFO_BINDING_MASK << INFO_BINDING_OFFSET)) | (binding << INFO_BINDING_OFFSET);
	}

	public void setLocal() {
		setBinding(STB_LOCAL);
	}

	public void setWeak() {
		setBinding(STB_WEAK);
	}

	public void setGlobal() {
		setBinding(STB_GLOBAL);
	}

	public void setTypeSection() {
		setType(STT_SECTION);
	}

	public void setTypeFunc() {
		setType(STT_FUNC);
	}

	public void setTypeObject() {
		setType(STT_OBJECT);
	}

	public void setSectionCommon() {
		this.sectionIndex = SHN_COMMON;
	}

	public boolean isLocal() {
		return getBinding() == STB_LOCAL;
	}

	public boolean isWeak() {
		return getBinding() == STB_WEAK;
	}

	public boolean isGlobal() {
		return getBinding() == STB_GLOBAL;
	}

	public boolean isTypeSection() {
		return getType() == STT_SECTION;
	}

	public boolean isTypeFunc() {
		return getType() == STT_FUNC;
	}

	public boolean isTypeFile() {
		return getType() == STT_FILE;
	}

	public boolean isTypeObject() {
		return getType() == STT_OBJECT;
	}

	public boolean isAbsolute() {
		return this.targetSection == null && this.sectionIndex == SHN_ABS;
	}

	@Override
	public void generateUsing(ElfGenerator gen) {
		gen.generateSymbolTableEntry(this);
	}

	public boolean equals(SymbolTableEntry other) {
		if (this.name == other.name) {
			return true;
		} else {
			return false;
		}
	}

	public boolean isUndefined() {
		return this.sectionIndex == SHN_UNDEF;
	}
}
