/*
 * Java
 *
 * Copyright 2013-2023 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.kf;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;

import com.is2t.elflw.allocator.SectionAllocator;
import com.is2t.elflw.input.AbstractElfLoader;
import com.is2t.elflw.nodes.AllocationSection;
import com.is2t.elflw.nodes.ElfRelocatableUnit;
import com.is2t.elflw.nodes.ElfUnit;
import com.is2t.elflw.nodes.ProgAllocationSection;
import com.is2t.elflw.nodes.ProgControlSection;
import com.is2t.elflw.nodes.ProgSection;
import com.is2t.elflw.nodes.RelocationSection;
import com.is2t.elflw.nodes.Section;
import com.is2t.elflw.nodes.SymbolTableEntry;
import com.is2t.elflw.nodes.SymbolTableSection;

import ej.annotation.Nullable;
import ej.bon.Constants;
import ej.kf.AlreadyLoadedFeatureException;
import ej.kf.Feature;
import ej.kf.IncompatibleFeatureException;
import ej.kf.InvalidFormatException;
import ej.kf.Kernel;

/**
 * Common <code>.fo</code> dynamic linker. It assumes that all allocated sections are used (no section shrink). See
 * subclasses for more details (different relocations policies).
 */
public abstract class CommonDynamicFeatureLoader implements IFeatureLoader {
	public static final int UNKNOWN_SYMBOL = InvalidFormatException.FIRST_SPECIFIC_ERROR;
	public static final int UNDEFINED_SYMBOL = InvalidFormatException.FIRST_SPECIFIC_ERROR + 1;
	public static final int RELOCATION_ERROR1 = InvalidFormatException.FIRST_SPECIFIC_ERROR + 2;
	public static final int RELOCATION_ERROR2 = InvalidFormatException.FIRST_SPECIFIC_ERROR + 3;
	public static final int UNEXPECTED_LOADFEATURE_ERROR = InvalidFormatException.FIRST_SPECIFIC_ERROR + 4;
	public static final int INVALID_ALIGNMENT = InvalidFormatException.FIRST_SPECIFIC_ERROR + 5;
	public static final int UNEXPECTED_COPY_ERROR = InvalidFormatException.FIRST_SPECIFIC_ERROR + 6;
	public static final int OOME_KERNEL_WORKING_BUFFER = InvalidFormatException.FIRST_SPECIFIC_ERROR + 7;
	public static final int TOO_MANY_DYNAMIC_FEATURES = InvalidFormatException.FIRST_SPECIFIC_ERROR + 8;
	public static final int CORRUPTED_CUSTOM_INSTALLED_FEATURE = InvalidFormatException.FIRST_SPECIFIC_ERROR + 9;
	public static final int INVALID_CUSTOM_INSTALLED_FEATURE_ALIGNMENT = InvalidFormatException.FIRST_SPECIFIC_ERROR
			+ 10;
	public static final int OVERLAP_ROM_AREA = InvalidFormatException.FIRST_SPECIFIC_ERROR + 11;
	public static final int OVERLAP_RAM_AREA = InvalidFormatException.FIRST_SPECIFIC_ERROR + 12;
	public static final int RAM_ADDRESS_CHANGED_CUSTOM_INSTALLED_FEATURE = InvalidFormatException.FIRST_SPECIFIC_ERROR
			+ 13;

	private static final String CONSTANT_PREFIX = "com.microej.runtime.kf.link.";

	protected static final String DEBUG_CONSTANT = CONSTANT_PREFIX + "debug.enabled";

	protected static final String RELOC_CHUNK_CONSTANT = CONSTANT_PREFIX + "chunk.relocations.count";

	public static final String TMP_TRANSFER_BUFFER_SIZE_CONSTANT = CONSTANT_PREFIX + "transferbuffer.size";

	/**
	 * Property suffix for architecture specific ELF symbol prefix.
	 * <p>
	 * See <code>init-kf-microjvm-dd/init.xml</code> script file.
	 */
	private static final String ARCH_SYMBOL_PREFIX_CONSTANT = CONSTANT_PREFIX + "symbol.prefix";

	private static final byte[] TYPES_HEADER_SYMBOL;
	private static final byte[] FEATURES_START_SYMBOL;
	private static final byte[] KERNEL_SIGNATURE_SYMBOL;

	static {
		String symbolPrefix = Constants.getString(ARCH_SYMBOL_PREFIX_CONSTANT);

		TYPES_HEADER_SYMBOL = (symbolPrefix + "_java_header_types_start").getBytes();
		FEATURES_START_SYMBOL = (symbolPrefix + "_java_features_start").getBytes();
		KERNEL_SIGNATURE_SYMBOL = (symbolPrefix + "_java_kernel_signature_start").getBytes();
	}

	// NOTE:
	// All working fields required for installing a Feature must be stored in the FoLoader instance.
	// This instance is correctly freed when installation is done (both in case of success or failure).

	/**
	 * Has a mean only within {@link #load(byte[], InputStream)} process.
	 */
	@Nullable
	protected FoLoader foLoader;

	protected int typeHeaderAddress;

	/**
	 * The list of allocated addresses for each Feature loaded in-place in the Kernel Working Buffer.
	 */
	private final HashMap<Feature, SectionAllocator> featureAllocationData;

	public CommonDynamicFeatureLoader() {
		this.featureAllocationData = new HashMap<>();
	}

	@Override
	public boolean canLoad(byte[] signature) {
		return AbstractElfLoader.canLoad(signature);
	}

	@Override
	public boolean canUnload() {
		return true;
	}

	@Override
	public Feature load(byte[] signature, InputStream is)
			throws IOException, InvalidFormatException, IncompatibleFeatureException, AlreadyLoadedFeatureException {
		// Load of a dynamic Feature supports two modes:
		// - in-place allocation and link in Kernel Working Buffer
		// - copy to custom ROM & RAM areas
		//
		// The ROM area is organized as following:
		// - Header (CustomInstalledFeatureStruct + Kernel UID)
		// - PROGBITS resources section
		// - PROGBITS code section
		// The RAM area is organized as following:
		// - NOBITS section1
		// - ...
		// - NOBITS sectionN
		//
		// NOTE: The Kernel UID is the Kernel on which this Feature is linked, which may
		// not be the same than the Kernel UID embedded in the Feature (the
		// Kernel on which the Feature has been built).

		NativeSectionAllocator sectionAllocator = new NativeSectionAllocator();
		FoLoader foLoader = this.foLoader = new FoLoader(sectionAllocator);
		int customFeatureHandle;
		Feature feature = null;
		try {
			// All sections allocated in this try-block will be released
			// if an exception is thrown.

			// - Load ELF unit
			foLoader.loadFO(signature, is);
			ElfRelocatableUnit unit = foLoader.getCurrentUnit();

			// - Check this Feature is allowed to be loaded.
			int abiVersion = unit.abiVersion;
			if (abiVersion != getSupportedABIVersion()) {
				throw new IncompatibleFeatureException(IncompatibleFeatureException.WRONG_FO_VERSION);
			}

			SymbolTableEntry expectedKernelSignatureEntry = getSymbolEntry(KERNEL_SIGNATURE_SYMBOL);

			// Check that the Feature can be linked on this Kernel.
			if (!KernelNatives.checkKernel(extractTargetBytes(expectedKernelSignatureEntry), unit.lsbAndNotMsb)) {
				throw new IncompatibleFeatureException(IncompatibleFeatureException.WRONG_KERNEL_UID);
			}

			// Here, the Feature can be linked without crashing the execution VM
			this.typeHeaderAddress = getSymbolAddress(TYPES_HEADER_SYMBOL);

			// - Resolve relocations
			Section[] sections = unit.getSections();
			resolveRelocations(sections);

			if (Constants.getBoolean(DEBUG_CONSTANT)) {
				dumpResolvedSymbols(unit);
			}

			// find root symbol
			int featuresHeaderStart = getSymbolAddress(FEATURES_START_SYMBOL);

			foLoader.transferToRelocatedSections(featuresHeaderStart);

			// add the linked Feature with no LLKERNEL callback in case of error
			feature = addLinkedFeature(featuresHeaderStart, foLoader.getCustomFeatureHandle(), true);
			assert feature != null;
		} finally {
			customFeatureHandle = foLoader.getCustomFeatureHandle();
			if (feature == null) {
				// Reclaim all allocated blocks if an error occurred
				sectionAllocator.reclaim();
				if (customFeatureHandle != Feature.FEATURE_CUSTOM_HANDLE_NONE) {
					// Free the custom allocated Feature
					DynamicLoaderNatives.freeFeature(customFeatureHandle);
				}
			}
			// Free memory working fields
			this.foLoader = null;
		}

		assert feature != null;
		if (customFeatureHandle == Feature.FEATURE_CUSTOM_HANDLE_NONE) {
			// A Feature installed in-place: keep the sectionAllocator
			// the allocated memory blocks will be reclaimed when the Feature will be
			// uninstalled
			this.featureAllocationData.put(feature, sectionAllocator);
		} else {
			// Feature installed in custom ROM & RAM areas
			sectionAllocator.reclaim();
		}

		return feature;
	}

	protected void resolveRelocations(Section[] sections) throws InvalidFormatException {
		for (Section section : sections) {
			if (section instanceof RelocationSection) {
				RelocationSection rs = (RelocationSection) section;
				Section targetSection = getRelocatedSection(rs);
				if (targetSection instanceof ProgAllocationSection) {
					applyRelocations(rs, (ProgAllocationSection) targetSection);
				}
				// else: most likely a relocation to a debug section (skipped)
			}
		}
	}

	/**
	 * @param featuresHeaderStart
	 *            the feature header start address if customFeatureHandle == {@link Feature#FEATURE_CUSTOM_HANDLE_NONE},
	 *            <code>0</code> otherwise
	 * @return the linked Feature or <code>null</code> if LLKERNEL callback returned to silently ignore the error
	 */
	@Nullable
	private Feature addLinkedFeature(int featuresHeaderStart, int customFeatureHandle, boolean noCallback)
			throws AlreadyLoadedFeatureException, InvalidFormatException, IncompatibleFeatureException {

		if (customFeatureHandle != Feature.FEATURE_CUSTOM_HANDLE_NONE) {
			// Check Feature has been linked against this Kernel
			int checkKernelResult = DynamicLoaderNatives.checkKernel(customFeatureHandle);
			switch (checkKernelResult) {
			case DynamicLoaderNatives.CHECK_KERNEL_IGNORE: {
				return null;
			}
			case DynamicLoaderNatives.CHECK_KERNEL_OK: {
				break;
			}
			case DynamicLoaderNatives.CHECK_KERNEL_CORRUPTED_CONTENT: {
				throw new InvalidFormatException(CORRUPTED_CUSTOM_INSTALLED_FEATURE);
			}
			case DynamicLoaderNatives.CHECK_KERNEL_INCOMPATIBLE_KERNEL_WRONG_UID: {
				throw new IncompatibleFeatureException(IncompatibleFeatureException.WRONG_KERNEL_UID);
			}
			case DynamicLoaderNatives.CHECK_KERNEL_INCOMPATIBLE_KERNEL_WRONG_ADDRESSES: {
				throw new IncompatibleFeatureException(IncompatibleFeatureException.WRONG_KERNEL_ADDRESSES);
			}
			case DynamicLoaderNatives.CHECK_KERNEL_RAM_ADDRESS_CHANGED: {
				throw new InvalidFormatException(RAM_ADDRESS_CHANGED_CUSTOM_INSTALLED_FEATURE);
			}
			default: {
				throw new AssertionError();
			}
			}
		}

		// Register the Feature to dynamically installed Features
		long result = DynamicLoaderNatives.addLinkedFeature(featuresHeaderStart, customFeatureHandle, noCallback);

		if (result >= DynamicLoaderNatives.LINK_FEATURE_ERROR_CODE_MIN
				&& result <= DynamicLoaderNatives.LINK_FEATURE_ERROR_CODE_MAX) {
			switch ((int) result) {
			case DynamicLoaderNatives.LINK_FEATURE_TOO_MANY_INSTALLED: {
				throw new InvalidFormatException(TOO_MANY_DYNAMIC_FEATURES);
			}
			case DynamicLoaderNatives.LINK_FEATURE_ALREADY_INSTALLED: {
				long previousFeatureSectionDescriptor = DynamicLoaderNatives.getAlreadyInstalledFeatureDescriptor();
				Feature f = Kernel.getFeatureByDescriptor(previousFeatureSectionDescriptor);
				assert f != null;
				throw new AlreadyLoadedFeatureException(f);
			}
			case DynamicLoaderNatives.LINK_FEATURE_ROM_OVERLAP: {
				throw new InvalidFormatException(OVERLAP_ROM_AREA);
			}
			case DynamicLoaderNatives.LINK_FEATURE_RAM_OVERLAP: {
				throw new InvalidFormatException(OVERLAP_RAM_AREA);
			}
			case DynamicLoaderNatives.LINK_FEATURE_IGNORE: {
				return null;
			}
			default: {
				throw new AssertionError();
			}
			}
		} else {
			long featureSectionDescriptor = result;
			return new Feature(this, featureSectionDescriptor, customFeatureHandle);
		}
	}

	@Override
	public void loadInstalledFeatures()
			throws InvalidFormatException, IncompatibleFeatureException, AlreadyLoadedFeatureException {
		int nbCustomInstalledFeatures = DynamicLoaderNatives.getCustomInstalledFeaturesCount();
		for (int i = 0; i < nbCustomInstalledFeatures; ++i) {
			int customFeatureHandle = DynamicLoaderNatives.getCustomInstalledFeatureHandle(i);
			assert customFeatureHandle != Feature.FEATURE_CUSTOM_HANDLE_NONE;
			Feature f = addLinkedFeature(0, customFeatureHandle, false);
			if (f != null) {
				Kernel.addInstalledFeature(f);
			}
		}
	}

	protected abstract void applyRelocations(RelocationSection rs, ProgAllocationSection relocatedSection)
			throws InvalidFormatException;

	/**
	 * Get the ABI version supported by this dynamic linker
	 *
	 * @return the ABI version supported by this dynamic linker
	 */
	protected abstract int getSupportedABIVersion();

	/**
	 * @param unit
	 * @throws InvalidFormatException
	 */
	private void dumpResolvedSymbols(ElfRelocatableUnit unit) throws InvalidFormatException {
		// dump resolved symbol values
		SymbolTableSection section = getSymbolTableSection(unit);
		int nbEntries = section.nbEntries();
		for (int i = nbEntries; --i > 0;) {// first symbol index is null
			SymbolTableEntry entry = AbstractElfLoader.loadSymbolTableEntry(unit, section, i);
			assert entry != null;
			if (entry.isLocal()) {
				break;
			}
			Section targetSection = entry.targetSection;
			if (targetSection != null) {
				int address = getSymbolAddress(entry);
				System.out.println(new String(entry.name) + " = " + address + " / " + toHexString(address));
			}
		}
	}

	private SymbolTableSection getSymbolTableSection(ElfUnit unit) {
		SymbolTableSection symbolTableSection = unit.symbolTableSection;
		assert symbolTableSection != null;
		return symbolTableSection;
	}

	/**
	 * Assume symbol targets a {@link ProgControlSection}
	 */
	private byte[] extractTargetBytes(SymbolTableEntry s) {
		byte[] targetBytes = new byte[s.size];
		Section targetSection = s.targetSection;
		assert targetSection != null;
		((ProgAllocationSection) targetSection).data.copyTo(s.value, targetBytes, 0, s.size);
		return targetBytes;
	}

	private String toHexString(int address) {
		StringBuffer sb = new StringBuffer();
		sb.append("0x");
		String str = Integer.toHexString(address);
		int length = str.length();
		for (int i = 8 - length; --i >= 0;) {
			sb.append('0');
		}
		sb.append(str);
		return sb.toString();
	}

	protected ProgSection getRelocatedSection(RelocationSection rs) {
		ProgSection targetSection = rs.targetSection;
		assert targetSection != null;
		return targetSection;
	}

	/**
	 * linear search: caller is strongly encouraged to store the result in a cache
	 */
	private SymbolTableEntry getSymbolEntry(byte[] symbolName) throws InvalidFormatException {
		// find the symbol entry
		ElfUnit unit = getFoLoader().getCurrentUnit();
		SymbolTableSection section = getSymbolTableSection(unit);
		int nbEntries = section.nbEntries();
		for (int i = nbEntries; --i > 0;) { // first symbol index is null
			SymbolTableEntry entry = AbstractElfLoader.loadSymbolTableEntry(unit, section, i);
			assert entry != null;
			if (equals(entry.name, symbolName)) {
				return entry;
			}
		}
		throw new InvalidFormatException(UNKNOWN_SYMBOL, new String(symbolName));
	}

	/**
	 * Get the current {@link FoLoader}. Can only be called during {@link #load(byte[], InputStream)}
	 *
	 * @return the current {@link FoLoader}
	 */
	protected FoLoader getFoLoader() {
		FoLoader foLoader = this.foLoader;
		assert foLoader != null;
		return foLoader;
	}

	/**
	 * @see #getSymbolEntry(byte[])
	 */
	private int getSymbolAddress(byte[] symbolName) throws InvalidFormatException {
		return getSymbolAddress(getSymbolEntry(symbolName));
	}

	private int getSymbolAddress(SymbolTableEntry symbolEntry) throws InvalidFormatException {
		if (symbolEntry.isUndefined()) {
			throw new InvalidFormatException(UNDEFINED_SYMBOL, new String(symbolEntry.name));
		} else if (symbolEntry.isAbsolute()) {
			return symbolEntry.value;
		} else {
			Section targetSection = symbolEntry.targetSection;
			assert targetSection != null;
			return (int) ((AllocationSection) targetSection).getRelocatedAddress() + symbolEntry.value;
		}
	}

	public static boolean equals(byte[] array1, byte[] array2) {
		if (array1.length != array2.length) {
			return false;
		}
		for (int i = array1.length; --i >= 0;) { // reverse order : most likely symbols that have the same end are
			// equals
			if (array1[i] != array2[i]) {
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean unload(Feature f) {
		boolean result = DynamicLoaderNatives.unlink(f);
		if (!result) {
			throw new AssertionError(); // sanity check
		}
		SectionAllocator sectionAllocator = this.featureAllocationData.remove(f);
		if (sectionAllocator != null) {
			// This Feature has been allocated in-place in Kernel Working Buffer
			// => free the allocated sections
			sectionAllocator.reclaim();
		}
		return true;
	}
}
