/*
 * Java
 *
 * Copyright 2021-2023 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.kf;

import com.is2t.elflw.allocator.MemoryBlock;
import com.is2t.elflw.input.soar.nodes.RelocationSectionSoar;
import com.is2t.elflw.nodes.DynamicAllocationSection;
import com.is2t.elflw.nodes.ProgAllocationSection;
import com.is2t.elflw.nodes.RelocationSection;
import com.is2t.elflw.nodes.Section;

import ej.bon.Constants;
import ej.kf.Feature;
import ej.kf.InvalidFormatException;

/**
 * Dynamic load and link of a {@link Feature}. Data is an ELF file generated by SOARv2 in "feature mode". This dynamic
 * linker has to resolve both Kernel and its relocations. Relocations are stored in a custom format, not using the ELF
 * standard for compaction.
 */
public class DynamicFeatureLoader2 extends CommonDynamicFeatureLoader {

	public static final int SUPPORTED_ABI_VERSION = 0x01;

	// NOTE:
	// Do not store working fields required for installing a Feature here.
	// See comment in superclass.

	public DynamicFeatureLoader2() {
	}

	@Override
	protected int getSupportedABIVersion() {
		return SUPPORTED_ABI_VERSION;
	}

	@Override
	protected void applyRelocations(RelocationSection rs, ProgAllocationSection relocatedSection)
			throws InvalidFormatException {
		assert rs instanceof RelocationSectionSoar;

		// SOARv2 format: there are only 2 PROGBITS sections (code and resources)
		long sectionCodeStartAddress = relocatedSection.getRelocatedAddress();
		long sectionResourceStartAddress = 0;
		long sectionStaticsStartAddress = 0;
		Section[] sections = getFoLoader().getCurrentUnit().getSections();
		assert sections != null;
		for (Section s : sections) {
			if (s instanceof ProgAllocationSection) {
				ProgAllocationSection ps = (ProgAllocationSection) s;
				long relocatedAddress = ps.getRelocatedAddress();
				if (relocatedAddress != sectionCodeStartAddress) {
					assert sectionResourceStartAddress == 0;
					sectionResourceStartAddress = relocatedAddress;
				}
			} else if (s instanceof DynamicAllocationSection) {
				DynamicAllocationSection ds = (DynamicAllocationSection) s;
				assert sectionStaticsStartAddress == 0;
				sectionStaticsStartAddress = ds.getRelocatedAddress();
			}
		}
		assert sectionResourceStartAddress != 0;

		MemoryBlock data = relocatedSection.data;
		long memoryBlockDescriptor = ((NativeMemoryBlock) data).memoryBlockDescriptor;

		byte[] relocationData = rs.data;
		int relocationsCount = Constants.getInt(RELOC_CHUNK_CONSTANT);
		int startIndex = 0;
		while (true) {
			int res = DynamicLoaderNatives.applyRelocations2(memoryBlockDescriptor, relocationData, startIndex,
					relocationsCount, this.typeHeaderAddress, sectionCodeStartAddress, sectionResourceStartAddress,
					sectionStaticsStartAddress);
			if (res == 0) {
				break; // all relocations successfully applied
			} else if (res == -1) {
				throw new AssertionError();
			} else if (res > 0) {
				throw new InvalidFormatException(res);
			} else {
				int previousIndex = startIndex;
				startIndex = -res;
				assert startIndex > previousIndex && startIndex < relocationData.length;

				// give the hand to other threads of same priority before applying a new chunk of relocations
				Thread.yield();
			}
		}
	}
}
