/*
 * Java
 *
 * Copyright 2022 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.kf;

import java.io.IOException;
import java.io.InputStream;

import com.is2t.elflw.allocator.MemoryBlock;
import com.is2t.elflw.input.ElfLoaderError;
import com.is2t.elflw.input.soar.ElfLoaderSoar;
import com.is2t.elflw.nodes.DynamicAllocationSection;
import com.is2t.elflw.nodes.ElfRelocatableUnit;
import com.is2t.elflw.nodes.ProgAllocationSection;
import com.is2t.elflw.nodes.RelocationSection;
import com.is2t.elflw.nodes.Section;
import com.is2t.elflw.support.SectionHeader;

import ej.annotation.Nullable;
import ej.bon.Constants;
import ej.kf.Feature;
import ej.kf.InvalidFormatException;

/**
 * <code>.fo</code> ELF loader.
 */
public class FoLoader extends ElfLoaderSoar {

	private static final int CUSTOM_HEADER_STRUCT_ALIGNMENT = 4;

	private final NativeSectionAllocator sectionAllocator;

	@Nullable
	protected ElfRelocatableUnit currentUnit;

	/**
	 * This array follows the {@link SectionHeader} array. Contains the 0 based relative addresses of
	 * {@link ProgAllocationSection} from ROM and {@link DynamicAllocationSection} from RAM.
	 */
	@Nullable
	private int[] sectionsRelativeAdresses;

	/**
	 * This array follows the {@link SectionHeader} array. Contains <code>true</code> if the section is a
	 * {@link ProgAllocationSection} that has {@link RelocationSection}, <code>false</code> otherwise.
	 */
	@Nullable
	private boolean[] hasRelocations;

	/**
	 * This array follows the {@link SectionHeader} array. Contains a 32bits crc for {@link ProgAllocationSection}.
	 */
	@Nullable
	private int[] crcs;

	private int customFeatureHandle;

	/**
	 * Only set when #customFeatureHandle is different than {@link Feature#FEATURE_CUSTOM_HANDLE_NONE}
	 */
	private int customHeaderSize;

	/**
	 * Only set when #customFeatureHandle is different than {@link Feature#FEATURE_CUSTOM_HANDLE_NONE}
	 */
	private int nbROMSections;

	/**
	 * Only set when #customFeatureHandle is different than {@link Feature#FEATURE_CUSTOM_HANDLE_NONE}
	 */
	private long startAddressROM;

	/**
	 * Only set when #customFeatureHandle is different than {@link Feature#FEATURE_CUSTOM_HANDLE_NONE}
	 */
	private long startAddressRAM;

	/**
	 * Only set when #customFeatureHandle is different than {@link Feature#FEATURE_CUSTOM_HANDLE_NONE}
	 */
	private int sizeROM;

	/**
	 * Only set when #customFeatureHandle is different than {@link Feature#FEATURE_CUSTOM_HANDLE_NONE}
	 */
	private int sizeRAM;

	public FoLoader(NativeSectionAllocator sectionAllocator) {
		this.sectionAllocator = sectionAllocator;
		this.customFeatureHandle = Feature.FEATURE_CUSTOM_HANDLE_NONE;
	}

	public void loadFO(byte[] elfSignature, InputStream is) throws IOException, InvalidFormatException {
		try {
			this.currentUnit = load(elfSignature, is);
		} catch (ElfLoaderError e) {
			// Not a compatible ELF file
			throw new InvalidFormatException(InvalidFormatException.LOADER_FORMAT_ERROR, e.getMessage());
		} catch (Throwable e) {
			throw unwrapInvalidFormatException(e);
		}
	}

	/**
	 * Wraps {@link InvalidFormatException} to a generic {@link RuntimeException} during ELF loading
	 *
	 * @param e
	 *            an {@link InvalidFormatException}
	 * @return a {@link RuntimeException} which cause is the exception given as parameter
	 * @see #unwrapInvalidFormatException(Throwable)
	 */
	private static RuntimeException wrapInvalidFormatException(InvalidFormatException e) {
		return new RuntimeException(e);
	}

	/**
	 * Unwraps an exception that occurred during ELF loading to a {@link InvalidFormatException}.
	 *
	 * @param e
	 *            a generic {@link Throwable} exception
	 * @return an {@link InvalidFormatException}
	 */
	private InvalidFormatException unwrapInvalidFormatException(Throwable e) {
		Throwable cause = e.getCause();
		if (cause instanceof InvalidFormatException) {
			return (InvalidFormatException) cause;
		} else {
			// any other error thrown (ArrayIndexOutOfBoundException...) is considered as a
			// corrupted .fo stream
			return new InvalidFormatException(InvalidFormatException.CORRUPTED_FO_STREAM, e.toString());
		}
	}

	@Override
	protected void sectionHeadersLoaded(SectionHeader[] sectionHeaders) {
		// First pass on sections to compute required ROM & RAM total sizes and sections
		// relative addresses.
		int nbSections = sectionHeaders.length;
		int[] sectionsRelativeAdresses = new int[nbSections];
		this.sectionsRelativeAdresses = sectionsRelativeAdresses;
		boolean[] hasRelocations = new boolean[nbSections];
		this.hasRelocations = hasRelocations;
		this.crcs = new int[nbSections];

		// Compute the number of sections which will have a checksum
		int nbROMSections = 0;
		for (int sectionIndex = 1; sectionIndex < nbSections; ++sectionIndex) {
			SectionHeader sectionHeader = sectionHeaders[sectionIndex];
			if ((sectionHeader.flags & SHF_ALLOC) == SHF_ALLOC && sectionHeader.type == SHT_PROGBITS) {
				++nbROMSections;
			}
		}

		int headerSize = DynamicLoaderNatives.getCustomInstalledFeatureStructSize(nbROMSections);
		int nextAddressROM = headerSize;
		int nextAddressRAM = 0;

		for (int sectionIndex = 1; sectionIndex < nbSections; ++sectionIndex) {
			SectionHeader sectionHeader = sectionHeaders[sectionIndex];
			if ((sectionHeader.flags & SHF_ALLOC) == SHF_ALLOC) {
				if (sectionHeader.type == SHT_PROGBITS) {
					nextAddressROM = align(nextAddressROM, sectionHeader.addralign);
					sectionsRelativeAdresses[sectionIndex] = nextAddressROM;
					nextAddressROM = nextAddressROM + sectionHeader.size;
				} else if (sectionHeader.type == SHT_NOBITS) {
					nextAddressRAM = align(nextAddressRAM, sectionHeader.addralign);
					sectionsRelativeAdresses[sectionIndex] = nextAddressRAM;
					nextAddressRAM = nextAddressRAM + sectionHeader.size;
				}
			}
			int relocatedSectionIndex = getRelocatedSection(sectionHeader);
			if (relocatedSectionIndex != 0) {
				hasRelocations[relocatedSectionIndex] = true;
			}
		}
		int sizeROM = nextAddressROM;
		int sizeRAM = nextAddressRAM;

		// - Allocate Feature ROM & RAM areas
		int customFeatureHandle = DynamicLoaderNatives.allocateFeature(sizeROM, sizeRAM);
		if (customFeatureHandle != Feature.FEATURE_CUSTOM_HANDLE_NONE) {
			this.customFeatureHandle = customFeatureHandle;
			this.startAddressROM = checkFeatureAddress(DynamicLoaderNatives.getFeatureAddressROM(customFeatureHandle));
			this.startAddressRAM = checkFeatureAddress(DynamicLoaderNatives.getFeatureAddressRAM(customFeatureHandle));
			this.sizeROM = sizeROM;
			this.sizeRAM = sizeRAM;
			this.customHeaderSize = headerSize;
			this.nbROMSections = nbROMSections;
		}
	}

	@Override
	protected ProgAllocationSection newProgAllocationSection(int sectionIndex, SectionHeader sectionHeader) {
		int alignment = sectionHeader.addralign;
		MemoryBlock allocatedBlock;
		MemoryBlock relocatedBlock;
		if (this.customFeatureHandle == Feature.FEATURE_CUSTOM_HANDLE_NONE) {
			// Installation in-place in Kernel Working Buffer
			try {
				allocatedBlock = allocateInKernelWorkingBuffer(alignment, sectionHeader.size);
			} catch (InvalidFormatException e) {
				throw wrapInvalidFormatException(e);
			}
			relocatedBlock = allocatedBlock;
		} else {
			// Installation to custom ROM area
			relocatedBlock = new MappedMemoryBlock(this.startAddressROM + getSectionRelativeAddress(sectionIndex),
					sectionHeader.size);

			if (hasRelocations(sectionIndex)) {
				// The section has relocations: the content is copied to Kernel Working
				// Buffer, then it will be transfered to ROM location after applying relocations
				try {
					allocatedBlock = allocateInKernelWorkingBuffer(alignment, sectionHeader.size);
				} catch (InvalidFormatException e) {
					throw wrapInvalidFormatException(e);
				}
			} else {
				// The section has no relocations, the content is directly transferred
				// to the relocatedBlock **without** allocation in the Kernel Working Buffer.
				// This is typically the case for the section that contains raw resources.
				allocatedBlock = relocatedBlock;
			}
		}
		return new ProgAllocationSection(sectionIndex, alignment, allocatedBlock, relocatedBlock);
	}

	private boolean hasRelocations(int sectionIndex) {
		boolean[] hasRelocations = this.hasRelocations;
		assert hasRelocations != null;
		return hasRelocations[sectionIndex];
	}

	private int getSectionRelativeAddress(int sectionIndex) {
		int[] sectionsRelativeAdresses = this.sectionsRelativeAdresses;
		assert sectionsRelativeAdresses != null;
		return sectionsRelativeAdresses[sectionIndex];
	}

	@Override
	protected DynamicAllocationSection newDynamicAllocationSection(int sectionIndex, SectionHeader sectionHeader) {
		int alignment = sectionHeader.addralign;
		MemoryBlock allocatedBlock;
		if (this.customFeatureHandle == Feature.FEATURE_CUSTOM_HANDLE_NONE) {
			// Installation in-place in Kernel Working Buffer
			// allocate NOBITS sections
			try {
				allocatedBlock = allocateInKernelWorkingBuffer(alignment, sectionHeader.size);
			} catch (InvalidFormatException e) {
				throw wrapInvalidFormatException(e);
			}
		} else {
			// Installation to custom RAM area
			allocatedBlock = new MappedMemoryBlock(this.startAddressRAM + getSectionRelativeAddress(sectionIndex),
					sectionHeader.size);
		}
		return new DynamicAllocationSection(alignment, allocatedBlock);
	}

	@Override
	protected void onCopyError() {
		throw wrapInvalidFormatException(newInvalidFormatCopyError());
	}

	@Override
	protected int getTransferBufferSize() {
		return Constants.getInt(CommonDynamicFeatureLoader.TMP_TRANSFER_BUFFER_SIZE_CONSTANT);
	}

	/**
	 * Allocates a new {@link MemoryBlock} in Kernel Working Buffer.
	 *
	 * @param alignment
	 *            the block alignment
	 * @param size
	 *            the block size in bytes
	 * @return the allocated block
	 * @throws InvalidFormatException
	 *             if the block could not be allocated
	 */
	public NativeMemoryBlock allocateInKernelWorkingBuffer(int alignment, int size) throws InvalidFormatException {
		NativeMemoryBlock allocatedBlock = this.sectionAllocator.allocate(alignment, size);
		if (allocatedBlock != null) {
			return allocatedBlock;
		} else {
			throw new InvalidFormatException(CommonDynamicFeatureLoader.OOME_KERNEL_WORKING_BUFFER,
					"(" + size + " bytes)");
		}
	}

	public static int align(int value, int alignment) {
		assert isPowerOfTwo(alignment);
		int alignmentMinusOne = alignment - 1;
		return (value + alignmentMinusOne) & ~alignmentMinusOne;
	}

	public static boolean isPowerOfTwo(int align) {
		// power of two: at most one bit set
		for (int i = 32; --i >= 0;) {
			if ((align & 1) == 1) {
				return (align & 0xFFFFFFFE) == 0;
			}
			align >>>= 1;
		}
		return false;
	}

	/**
	 * Require {@link #loadFO(byte[], InputStream)} called before.
	 *
	 * @return the custom allocated Feature handle, or {@link Feature#FEATURE_CUSTOM_HANDLE_NONE} if none
	 */
	public int getCustomFeatureHandle() {
		return this.customFeatureHandle;
	}

	@Override
	protected void transferSectionDataToMemoryBlock(ProgAllocationSection section, int[] absolutePtrRef, InputStream is)
			throws IOException {
		startCRC();
		super.transferSectionDataToMemoryBlock(section, absolutePtrRef, is);
		endCRC(section);
	}

	private void startCRC() {
		DynamicLoaderNatives.copyToROMCRCStart();
	}

	private void endCRC(ProgAllocationSection section) {
		int[] crcs = this.crcs;
		assert crcs != null;
		crcs[section.getSectionIndex()] = DynamicLoaderNatives.copyToROMCRCEnd();
	}

	public void transferToRelocatedSections(int featuresHeaderStart) throws InvalidFormatException {
		int customFeatureHandle = this.customFeatureHandle;
		if (customFeatureHandle != Feature.FEATURE_CUSTOM_HANDLE_NONE) {
			// Transfer linked content to ROM
			// This may take a long time, especially in case of In System Programming to
			// Flash Memory

			// Transfer PROGBITS sections
			ElfRelocatableUnit currentUnit = getCurrentUnit();
			Section[] sections = currentUnit.getSections();
			for (Section section : sections) {
				if (section instanceof ProgAllocationSection) {
					ProgAllocationSection progSection = (ProgAllocationSection) section;
					long relocatedAddress = progSection.getRelocatedAddress();
					long address = progSection.getMemoryBlock().getAddress();
					if (address != relocatedAddress) {
						startCRC();
						int size = progSection.size();
						copyToROM(relocatedAddress, address, size);
						endCRC(progSection);
					} else {
						// the content has already been transfered to the relocated block during
						// ELF loading (resources section with no relocations)
					}
				}
			}

			// Initialize the header
			int headerSize = this.customHeaderSize;
			int nbROMSections = this.nbROMSections;
			NativeMemoryBlock headerBlock = allocateInKernelWorkingBuffer(CUSTOM_HEADER_STRUCT_ALIGNMENT, headerSize);
			DynamicLoaderNatives.initializeCustomInstalledFeatureStruct(headerBlock.memoryBlockDescriptor,
					customFeatureHandle, featuresHeaderStart, this.sizeROM, this.sizeRAM, nbROMSections);

			int crcROMSectionIndex = 0;
			for (Section section : sections) {
				if (section instanceof ProgAllocationSection) {
					ProgAllocationSection progSection = (ProgAllocationSection) section;
					int sectionIndex = progSection.getSectionIndex();
					int relativeAddressROM = getSectionRelativeAddress(sectionIndex);
					int size = progSection.size();
					int crc = getCRC(sectionIndex);
					DynamicLoaderNatives.initializeCustomInstalledFeatureStructCRCSection(
							headerBlock.memoryBlockDescriptor, crcROMSectionIndex, relativeAddressROM, size, crc);
					++crcROMSectionIndex;
				}
			}
			assert crcROMSectionIndex == nbROMSections;
			DynamicLoaderNatives.finalizeCustomInstalledFeatureStruct(headerBlock.memoryBlockDescriptor);

			copyToROM(this.startAddressROM, headerBlock.getAddress(), headerSize);

			// The last call to notify all bytes have been transfered to ROM
			checkCopyToROM(DynamicLoaderNatives.flushCopyToROM());
		}

	}

	private int getCRC(int sectionIndex) {
		int[] crcs = this.crcs;
		assert crcs != null;
		return crcs[sectionIndex];
	}

	private void copyToROM(long relocatedAddress, long address, int size) throws InvalidFormatException {
		checkCopyToROM(DynamicLoaderNatives.copyKernelWorkingBufferToROM(relocatedAddress, address, size));
	}

	public static void checkCopyToROM(boolean result) throws InvalidFormatException {
		if (!result) {
			throw newInvalidFormatCopyError();
		}
	}

	public static long checkFeatureAddress(long startAddressROMOrRAM) throws ElfLoaderError {
		if (startAddressROMOrRAM == 0) {
			// the address does not respect the memory alignment
			throw wrapInvalidFormatException(
					new InvalidFormatException(CommonDynamicFeatureLoader.INVALID_CUSTOM_INSTALLED_FEATURE_ALIGNMENT));
		}
		return startAddressROMOrRAM;
	}

	private static InvalidFormatException newInvalidFormatCopyError() {
		return new InvalidFormatException(CommonDynamicFeatureLoader.UNEXPECTED_COPY_ERROR);
	}

	public ElfRelocatableUnit getCurrentUnit() {
		ElfRelocatableUnit currentUnit = this.currentUnit;
		assert currentUnit != null;
		return currentUnit;
	}
}
