/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2021-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;

import ej.sni.NativeIOException;

/**
 * DNS Native methods.
 */
public class DNSNatives {

	static {
		ChannelNatives.initializeNativePart();
	}

	/**
	 * Gets the host name associated to the input IP address {@code address}.
	 * 
	 * The given IP address is stored in network byte order (highest order byte of the address stored in
	 * {@code address[0]}). Example: Assume that the given IP address is 192.168.4.10, the content of the input
	 * {@code address} buffer will be: {@code address[0]=192; address[1]=168; address[2]=4; address[3]=10</code>}.
	 *
	 * <p>
	 * When this method returns, the resolved host name is stored in the {@code hostname} buffer.
	 * 
	 * @param address
	 *            the IP address whose the host name will be got
	 * @param addressLength
	 *            the IP address size (4 for IPv4 address or 16 for IPv6 address)
	 * @param hostname
	 *            the output buffer into which the resolved host name will be stored. (The host name must be a
	 *            null-terminated string)
	 * @param hostnameLength
	 *            the {@code hostname} buffer length.
	 * @return the host name length.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes.
	 */
	public static native int getHostByAddr(byte[] address, int addressLength, byte[] hostname, int hostnameLength)
			throws NativeIOException;

	/**
	 * Gets the IP address at the {@code index} position in the resolved IP addresses list of the given host name
	 * {@code hostname}.
	 * 
	 * <p>
	 * The buffer {@code hostname} contains the host name whose the IP address will be got.
	 * 
	 * When this method returns, the IP address is stored in the {@code address} buffer.
	 * 
	 * The IP address will be stored in network byte order (highest order byte of the address stored in
	 * {@code address[0]}). Example: Assume that the IP address is 192.168.4.10, the content of the output
	 * {@code address} buffer will be: {@code address[0]=192; address[1]=168; address[2]=4; address[3]=10}.
	 * 
	 * @param index
	 *            the index of the IP address to get.
	 * @param hostname
	 *            the input host name (null-terminated string).
	 * @param hostnameLength
	 *            the host name length. (The length does not include the terminating null byte).
	 * @param address
	 *            the output buffer into which the IP address will be stored.
	 * @param addressLength
	 *            the output IP address buffer length.
	 * @return the IP address size in bytes (4 for IPv4 address or 16 for IPv6 address).
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes.
	 */
	public static native int getHostByNameAt(int index, byte[] hostname, int hostnameLength, byte[] address,
			int addressLength) throws NativeIOException;

	/**
	 * Returns the number of host addresses associated with the given host name {@code hostname}.
	 *
	 * @param hostname
	 *            the host name buffer (null-terminated string).
	 * @param hostnameLength
	 *            the host name length. (The length does not include the terminating null byte).
	 * @return the number of addresses associated with the host name {@code hostname}.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes.
	 */
	public static native int getHostByNameCount(byte[] hostname, int hostnameLength) throws NativeIOException;
}
