/*
 * Copyright 2017-2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security;

import java.security.KeyFactorySpi;
import java.security.KeyPairGeneratorSpi;
import java.security.MessageDigestSpi;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandomSpi;
import java.security.SignatureSpi;

import javax.crypto.CipherSpi;
import javax.crypto.MacSpi;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKeyFactorySpi;

import com.is2t.support.security.cipher.NativeCipherSpi;
import com.is2t.support.security.cipher.NativeRSACipherSpi;
import com.is2t.support.security.cipher.PBES2CipherSpi;
import com.is2t.support.security.cipher.PKCS12PBECipherSpi;
import com.is2t.support.security.digest.NativeMessageDigestSpi;
import com.is2t.support.security.keyfactory.NativeKeyFactorySpi;
import com.is2t.support.security.mac.NativeMacSpi;
import com.is2t.support.security.securerandom.NativeSecureRandomSpi;
import com.is2t.support.security.signature.NativeSignatureSpi;
import com.microej.support.security.keypairgen.NativeKeyPairGeneratorSpi;
import com.microej.support.security.pbekey.PBEKeyFactory;
import com.microej.support.security.secretkey.NativeSecretKeyFactorySpi;

import ej.bon.ByteArray;
import ej.sni.SNI;

/**
 *
 */
public class NativeSecurityProvider implements SecurityProvider {

	/**
	 * Length of the string to provide for the message digest algorithm name.
	 */
	public static final int MESSAGE_DIGEST_ALGORITHM_NAME_SIZE = 16;

	private CipherSpi getNativeCipher(String transformation) {
		byte[] cString = new byte[transformation.length() + 1];// add 1 for the '\0'
		SNI.toCString(transformation, cString);

		int cipherDescriptionSize = NativeCipherSpi.CIPHER_DESCRIPTION_SIZE;

		byte[] cipherDescription = new byte[cipherDescriptionSize];
		int nativeTransformationId = -1;

		nativeTransformationId = NativeCipherSpi.nativeGetTransformationDescription(cString, cipherDescription);
		if (nativeTransformationId != -1) {
			int blockSize = ByteArray.readInt(cipherDescription, 0);
			int unitBytes = ByteArray.readInt(cipherDescription, ByteArray.INT_SIZE);
			int cipherMode = ByteArray.readInt(cipherDescription, ByteArray.INT_SIZE * 2);

			return new NativeCipherSpi(nativeTransformationId, blockSize, unitBytes, cipherMode);
		}

		return null;
	}

	private CipherSpi getNativeRSACipher(String transformation) {
		byte[] cString = new byte[transformation.length() + 1];// add 1 for the '\0'
		SNI.toCString(transformation, cString);

		int cipherDescriptionSize = NativeRSACipherSpi.CIPHER_DESCRIPTION_SIZE;

		byte[] cipherDescription = new byte[cipherDescriptionSize];
		int nativeTransformationId = -1;

		nativeTransformationId = NativeRSACipherSpi.nativeGetTransformationDescription(cString, cipherDescription);
		if (nativeTransformationId != -1) {
			int paddingType = ByteArray.readInt(cipherDescription, 0);
			int oaepHashAlgorithm = ByteArray.readInt(cipherDescription, ByteArray.INT_SIZE);

			return new NativeRSACipherSpi(nativeTransformationId, paddingType, oaepHashAlgorithm);
		}

		return null;
	}

	private CipherSpi getPKCS12PBECipher(String transformation)
			throws NoSuchAlgorithmException, NoSuchPaddingException {
		// The PKCS12PBECipherSpi is not a native cipher, but relies instead of the following components:
		// - A native message digest algorithm
		// - A native symmetric cipher

		// Let the PKCS12PBECipherSpi return the MD algorithm and symmetric cipher transformation from an input PKCS12
		// PBE cipher transformation
		String mdAlgorithm = PKCS12PBECipherSpi.getMdAlgorithm(transformation);
		String cipherTransformation = PKCS12PBECipherSpi.getCipherTransformation(transformation);

		// If the translation is successful
		if ((mdAlgorithm != null) && (cipherTransformation != null)) {
			NativeMessageDigestSpi messageDigest = (NativeMessageDigestSpi) getMessageDigestSpi(mdAlgorithm);
			NativeCipherSpi nativeCipher = (NativeCipherSpi) getNativeCipher(cipherTransformation);

			// If the native message digest algorithm and symmetric cipher are available, then the PKCS12 PBE cipher can
			// use them
			if ((messageDigest != null) && (nativeCipher != null)) {
				return new PKCS12PBECipherSpi(messageDigest, nativeCipher);
			}
		}

		return null;
	}

	private CipherSpi getPBES2Cipher(String transformation) throws NoSuchAlgorithmException, NoSuchPaddingException {
		// The PBES2CipherSpi is not a native cipher, but relies instead of the following components:
		// - A native password-based key derivation algorithm
		// - A native symmetric cipher

		// Let the PBES2CipherSpi return the KDF algorithm and symmetric cipher transformation from an input PBES2
		// cipher transformation
		String kdfAlgorithm = PBES2CipherSpi.getKdfAlgorithm(transformation);
		String cipherTransformation = PBES2CipherSpi.getCipherTransformation(transformation);

		// If the translation is successful
		if ((kdfAlgorithm != null) && (cipherTransformation != null)) {
			NativeSecretKeyFactorySpi secretKeyFactory = (NativeSecretKeyFactorySpi) getSecretKeyFactorySpi(
					kdfAlgorithm);
			NativeCipherSpi nativeCipher = (NativeCipherSpi) getNativeCipher(cipherTransformation);

			// If the native message digest algorithm and symmetric cipher are available, then the PBES2 cipher can use
			// them
			if ((secretKeyFactory != null) && (nativeCipher != null)) {
				return new PBES2CipherSpi(secretKeyFactory, nativeCipher);
			}
		}

		return null;
	}

	@Override
	public CipherSpi getCipherSpi(String transformation) throws NoSuchAlgorithmException, NoSuchPaddingException {
		CipherSpi cipherSpi;

		cipherSpi = getNativeCipher(transformation);
		if (cipherSpi != null) {
			return cipherSpi;
		}

		cipherSpi = getNativeRSACipher(transformation);
		if (cipherSpi != null) {
			return cipherSpi;
		}

		cipherSpi = getPKCS12PBECipher(transformation);
		if (cipherSpi != null) {
			return cipherSpi;
		}

		cipherSpi = getPBES2Cipher(transformation);
		if (cipherSpi != null) {
			return cipherSpi;
		}

		throw new NoSuchAlgorithmException("Cipher transformation + " + transformation + " not supported");
	}

	@Override
	public SignatureSpi getSignatureSpi(String algorithm) {

		byte[] cAlgorithm = new byte[algorithm.length() + 1];// add 1 for the '\0'
		SNI.toCString(algorithm, cAlgorithm);

		byte[] cMessageDigestAlgorithmName = new byte[MESSAGE_DIGEST_ALGORITHM_NAME_SIZE];

		int nativeAlgorithmId = NativeSignatureSpi.nativeGetAlgorithmDescription(cAlgorithm,
				cMessageDigestAlgorithmName, cMessageDigestAlgorithmName.length);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		String messageDigestAlgorithmName = SNI.toJavaString(cMessageDigestAlgorithmName);

		return new NativeSignatureSpi(nativeAlgorithmId, messageDigestAlgorithmName);
	}

	@Override
	public MessageDigestSpi getMessageDigestSpi(String algorithm) {

		byte[] cString = new byte[algorithm.length() + 1];// add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		byte[] algorithmDescription = new byte[NativeMessageDigestSpi.ALGORITHM_DESCRIPTION_SIZE];

		int nativeAlgorithmId = NativeMessageDigestSpi.nativeGetAlgorithmDescription(cString, algorithmDescription);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		int digestLength = ByteArray.readInt(algorithmDescription, 0);

		return new NativeMessageDigestSpi(nativeAlgorithmId, digestLength);
	}

	@Override
	public MacSpi getMacSpi(String algorithm) {

		byte[] cString = new byte[algorithm.length() + 1];// add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		byte[] algorithmDescription = new byte[NativeMacSpi.ALGORITHM_DESCRIPTION_SIZE];

		int nativeAlgorithmId = NativeMacSpi.nativeGetAlgorithmDescription(cString, algorithmDescription);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		int macLength = ByteArray.readInt(algorithmDescription, 0);

		return new NativeMacSpi(nativeAlgorithmId, macLength);
	}

	@Override
	public SecureRandomSpi getSecureRandomSpi() {
		return new NativeSecureRandomSpi();
	}

	@Override
	public KeyFactorySpi getKeyFactorySpi(String algorithm) {
		byte[] cString = new byte[algorithm.length() + 1]; // add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		int nativeAlgorithmId = NativeKeyFactorySpi.nativeGetAlgorithmDescription(cString);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		return new NativeKeyFactorySpi(nativeAlgorithmId, algorithm);
	}

	@Override
	public KeyPairGeneratorSpi getKeyPairGeneratorSpi(String algorithm) throws NoSuchAlgorithmException {
		// check that the algorithm is supported by the native implementation
		byte[] cString = new byte[algorithm.length() + 1]; // add 1 for the '\0'
		SNI.toCString(algorithm, cString);
		int nativeAlgorithmId = NativeKeyPairGeneratorSpi.nativeGetAlgorithm(cString);
		if (nativeAlgorithmId == -1) {
			throw new NoSuchAlgorithmException(algorithm + " not found.");
		}

		return new NativeKeyPairGeneratorSpi(nativeAlgorithmId, algorithm);
	}

	@Override
	public SecretKeyFactorySpi getSecretKeyFactorySpi(String algorithm) throws NoSuchAlgorithmException {
		byte[] cString = new byte[algorithm.length() + 1]; // add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		// Search the requested transformation through the NativeSecretKeyFactorySpi
		int nativeAlgorithmId = NativeSecretKeyFactorySpi.nativeGetAlgorithm(cString);
		if (nativeAlgorithmId != -1) {
			return new NativeSecretKeyFactorySpi(nativeAlgorithmId, algorithm);
		} else if (PBEKeyFactory.isAlgorithmSupported(algorithm)) { // Search the requested transformation through the
																	// PBEKeyFactory, which is not a native key factory
			return new PBEKeyFactory(algorithm);
		} else {
			throw new NoSuchAlgorithmException(algorithm + " not found.");
		}
	}

}
