/*
 * Copyright 2023 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.cipher;

import java.security.ProviderException;

import com.is2t.support.security.NativePrivateKey;
import com.is2t.support.security.NativePublicKey;

import ej.bon.ByteArray;
import ej.sni.NativeException;
import ej.sni.NativeResource;

/**
 * Native implementation of the RSA cipher.
 */
public class NativeRSACipherSpi extends AbstractRSACipherSpi {

	/**
	 * @see #nativeGetTransformationDescription(byte[], byte[])
	 */
	public static final int CIPHER_DESCRIPTION_SIZE = ByteArray.INT_SIZE * 2;

	/**
	 * Native identifier for the cipher. Initialized during engine init.
	 */
	private int nativeId;

	/**
	 * Native transformation id of the cipher.
	 */
	private final int nativeTransformationId;

	/**
	 * Custom constructor.
	 *
	 * @param nativeTransformationId
	 *            transformation id received from the native world.
	 * @param paddingType
	 *            padding type constant.
	 * @param oaepHashAlgorithm
	 *            OAEP hash algorithm (if used).
	 */
	public NativeRSACipherSpi(int nativeTransformationId, int paddingType, int oaepHashAlgorithm) {
		super(paddingType, oaepHashAlgorithm);
		this.nativeId = -1;
		this.nativeTransformationId = nativeTransformationId;
	}

	@Override
	protected void init(NativePublicKey publicKey, NativePrivateKey privateKey, int padding, int hash) {
		try {
			if (this.nativeId != -1) {
				// close and reset the existing resource
				nativeClose(this.nativeTransformationId, this.nativeId);
				this.nativeId = -1;
			}

			boolean decrypting = false;
			int keyNativeId = 0;

			if (publicKey != null) {
				if (privateKey == null) {
					keyNativeId = publicKey.getNativeId();
				} else {
					throw new IllegalArgumentException("Bad key input");
				}
			} else if (privateKey != null) {
				keyNativeId = privateKey.getNativeId();
				decrypting = true;
			} else {
				throw new IllegalArgumentException("Bad key input");
			}

			int transformationId = this.nativeTransformationId;

			try {
				this.nativeId = nativeInit(transformationId, decrypting, keyNativeId, padding, hash);
				NativeResource.closeOnGC(this.nativeId, nativeGetCloseId(transformationId), this);
			} catch (NativeException e) {
				// An error occurs during initialization, maybe there is no space left to
				// allocate new native resource.
				// Trigger a new fresh GC and release new potential useless resources and try
				// the initialization again!
				System.gc();
				this.nativeId = nativeInit(transformationId, decrypting, keyNativeId, padding, hash);
				NativeResource.closeOnGC(this.nativeId, nativeGetCloseId(transformationId), this);
			}
		} catch (NativeException e) {
			throw new ProviderException(e);
		}
	}

	@Override
	protected int encrypt(byte[] in, int inOffset, int inLength, byte[] output, int outputOffset) {
		return nativeEncrypt(this.nativeTransformationId, this.nativeId, in, inOffset, inLength, output, outputOffset);
	}

	@Override
	protected int decrypt(byte[] in, int inOffset, int inLength, byte[] output, int outputOffset) {
		return nativeDecrypt(this.nativeTransformationId, this.nativeId, in, inOffset, inLength, output, outputOffset);
	}

	/**
	 * Gets for the given transformation the cipher description.
	 *
	 * @param transformation
	 *            Null terminated string that describes the transformation
	 * @param description
	 *            must be filled-in with:
	 *            <ul>
	 *            <li>[0-3]: paddingType: one of the <code>PAD_*</code> values defined in
	 *            {@link AbstractRSACipherSpi}</li>
	 *            <li>[4-7]: oaepHashAlgorithm: one of the <code>OAEP_HASH_*</code> values defined in
	 *            {@link AbstractRSACipherSpi}</li>
	 *            </ul>
	 * @return transformation ID on success or -1 on error.
	 */
	public static native int nativeGetTransformationDescription(byte[] transformation, byte[] description);

	/**
	 * Initializes a Cipher resource.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param decrypting
	 *            <code>true</code> for decrypting, <code>false</code> for encryting.
	 * @param keyNativeId
	 *            ID of the key used by the cipher
	 * @param padding
	 *            padding mode used by the cipher
	 * @param hash
	 *            hash algorithm used by the cipher
	 *
	 * @return the nativeId of the newly initialized resource
	 * @throws NativeException
	 *             on error.
	 */
	private static native int nativeInit(int nativeTransformationId, boolean decrypting, int keyNativeId, int padding,
			int hash) throws NativeException;

	/**
	 * Encrypts the message contained in <code>buffer</code>.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @param buffer
	 *            the buffer containing the plaintext message to encrypt
	 * @param bufferOffset
	 *            the buffer offset
	 * @param len
	 *            the buffer length
	 * @param output
	 *            the output buffer containing the encrypted message
	 * @param outputOffset
	 *            the output offset
	 * @return the length of the buffer
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeEncrypt(int nativeTransformationId, int nativeId, byte[] buffer, int bufferOffset,
			int len, byte[] output, int outputOffset) throws NativeException;

	/**
	 * Decrypts the message contained in <code>buffer</code>.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @param buffer
	 *            the buffer containing the message to decrypt
	 * @param bufferOffset
	 *            the buffer offset
	 * @param len
	 *            the buffer length
	 * @param output
	 *            the output buffer containing the plaintext message
	 * @param outputOffset
	 *            the output offset
	 * @return the length of the buffer
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeDecrypt(int nativeTransformationId, int nativeId, byte[] buffer, int bufferOffset,
			int len, byte[] output, int outputOffset) throws NativeException;

	/**
	 * Closes the resource related to the native ID
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @throws NativeException
	 *             on error
	 */
	private static native void nativeClose(int nativeTransformationId, int nativeId) throws NativeException;

	/**
	 * Gets the id of the native close function.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetCloseId(int nativeTransformationId) throws NativeException;

}
