/*
 * Copyright 2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.cipher;

import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;

import javax.crypto.NoSuchPaddingException;

import com.microej.support.security.secretkey.NativeSecretKeyFactorySpi;

/**
 * Implementation of the PBES2 cipher.
 */
public class PBES2CipherSpi extends AbstractPBES2CipherSpi {

	private static final Map<String, String> kdfAlgorithms;
	private static final Map<String, String> cipherTransformations;

	static {
		kdfAlgorithms = new HashMap<>();
		// Only PBES2 ciphers without HMAC output truncation and with AES_128 cipher are supported.
		kdfAlgorithms.put("PBEWithHmacSHA1AndAES_128".toUpperCase(), "PBKDF2WithHmacSHA1");
		kdfAlgorithms.put("PBEWithHmacSHA224AndAES_128".toUpperCase(), "PBKDF2WithHmacSHA224");
		kdfAlgorithms.put("PBEWithHmacSHA256AndAES_128".toUpperCase(), "PBKDF2WithHmacSHA256");
		kdfAlgorithms.put("PBEWithHmacSHA384AndAES_128".toUpperCase(), "PBKDF2WithHmacSHA384");
		kdfAlgorithms.put("PBEWithHmacSHA512AndAES_128".toUpperCase(), "PBKDF2WithHmacSHA512");

		cipherTransformations = new HashMap<>();
		// Only PBES2 ciphers without HMAC output truncation and with AES_128 cipher are supported.
		// PBES2 requires PKCS5Padding for the underlying AES cipher, but the padding is managed in the foundation
		// library, therefore we can use a native AES cipher without padding
		cipherTransformations.put("PBEWithHmacSHA1AndAES_128".toUpperCase(), "AES/CBC/NoPadding");
		cipherTransformations.put("PBEWithHmacSHA224AndAES_128".toUpperCase(), "AES/CBC/NoPadding");
		cipherTransformations.put("PBEWithHmacSHA256AndAES_128".toUpperCase(), "AES/CBC/NoPadding");
		cipherTransformations.put("PBEWithHmacSHA384AndAES_128".toUpperCase(), "AES/CBC/NoPadding");
		cipherTransformations.put("PBEWithHmacSHA512AndAES_128".toUpperCase(), "AES/CBC/NoPadding");
	}

	/**
	 * Custom constructor.
	 *
	 * @param secretKeyFactory
	 *            the secret key factory that can be used by the PBE cipher.
	 * @param cipher
	 *            the symmetric cipher that can be used by the PBE cipher.
	 * @throws NoSuchAlgorithmException
	 *             if the algorithm is not supported.
	 * @throws NoSuchPaddingException
	 *             if the padding is not supported.
	 */
	public PBES2CipherSpi(NativeSecretKeyFactorySpi secretKeyFactory, NativeCipherSpi cipher)
			throws NoSuchAlgorithmException, NoSuchPaddingException {
		super(secretKeyFactory, cipher);
	}

	/**
	 * Returns the corresponding key derivation algorithm that can be used by a PBE cipher.
	 *
	 * @param transformation
	 *            PBE cipher transformation
	 * @return the corresponding key derivation algorithm that can be used by the PBE cipher.
	 */
	public static String getKdfAlgorithm(String transformation) {
		if (transformation != null) {
			return kdfAlgorithms.get(transformation.toUpperCase());
		} else {
			return null;
		}
	}

	/**
	 * Returns the corresponding symmetric cipher transformation that can be used by a PBE cipher.
	 *
	 * @param transformation
	 *            PBE cipher transformation
	 * @return the corresponding symmetric cipher transformation that can be used by the PBE cipher.
	 */
	public static String getCipherTransformation(String transformation) {
		if (transformation != null) {
			return cipherTransformations.get(transformation.toUpperCase());
		} else {
			return null;
		}
	}

}
