/*
 * Copyright 2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.cipher;

import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;

import javax.crypto.NoSuchPaddingException;

import com.is2t.support.security.digest.NativeMessageDigestSpi;

/**
 * Implementation of the PKCS12 PBE cipher.
 */
public class PKCS12PBECipherSpi extends AbstractPKCS12PBECipherSpi {

	private static final Map<String, String> mdAlgorithms;
	private static final Map<String, String> cipherTransformations;

	static {
		mdAlgorithms = new HashMap<>();
		// Only PKCS12 PBE ciphers with 3DES cipher are supported.
		mdAlgorithms.put("PBEWithSHA1AndDESede".toUpperCase(), "SHA-1");

		cipherTransformations = new HashMap<>();
		// Only PKCS12 PBE ciphers with 3DES cipher are supported.
		// PKCS12 PBE requires PKCS5Padding for the underlying DES cipher, but the padding is managed in the foundation
		// library, therefore we can use a native DES cipher without padding
		cipherTransformations.put("PBEWithSHA1AndDESede".toUpperCase(), "DESede/CBC/NoPadding");
	}

	/**
	 * Custom constructor.
	 *
	 * @param messageDigest
	 *            the message digest that can be used by the PBE cipher.
	 * @param cipher
	 *            the symmetric cipher that can be used by the PBE cipher.
	 * @throws NoSuchAlgorithmException
	 *             if the algorithm is not supported.
	 * @throws NoSuchPaddingException
	 *             if the padding is not supported.
	 */
	public PKCS12PBECipherSpi(NativeMessageDigestSpi messageDigest, NativeCipherSpi cipher)
			throws NoSuchAlgorithmException, NoSuchPaddingException {
		super(messageDigest, cipher);
	}

	/**
	 * Returns the corresponding message digest algorithm that can be used by a PBE cipher.
	 *
	 * @param transformation
	 *            PBE cipher transformation
	 * @return the corresponding message digest algorithm that can be used by the PBE cipher.
	 */
	public static String getMdAlgorithm(String transformation) {
		if (transformation != null) {
			return mdAlgorithms.get(transformation.toUpperCase());
		} else {
			return null;
		}
	}

	/**
	 * Returns the corresponding symmetric cipher transformation that can be used by a PBE cipher.
	 *
	 * @param transformation
	 *            PBE cipher transformation
	 * @return the corresponding symmetric cipher transformation that can be used by the PBE cipher.
	 */
	public static String getCipherTransformation(String transformation) {
		if (transformation != null) {
			return cipherTransformations.get(transformation.toUpperCase());
		} else {
			return null;
		}
	}

}
