/*
 * Copyright (c) 2009, 2017, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2017-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.signature;

import java.security.InvalidKeyException;
import java.security.PrivateKey;
import java.security.PublicKey;

import com.is2t.support.security.NativePrivateKey;
import com.is2t.support.security.NativePublicKey;

import ej.sni.NativeException;

/**
 *
 */
public class NativeSignatureSpi extends AbstractSignatureSpi {

	private final int nativeAlgorithmId;

	private NativePublicKey publicKey;
	private NativePrivateKey privateKey;

	/**
	 * Default max signature size. Often equals to private key size / 8. 512 can handle signature for key of size 4096
	 * bits.
	 */
	private static final int DEFAULT_MAX_SIGNATURE_SIZE = 512;

	/**
	 * Property to configure the max signature signature size.
	 */
	private static final String MAX_SIGNATURE_SIZE_PROPERTY = "com.microej.security.signature.max.size";

	public NativeSignatureSpi(int nativeAlgorithmId, String messageDigestAlgorithmName) {
		super(messageDigestAlgorithmName);
		this.nativeAlgorithmId = nativeAlgorithmId;
	}

	@Override
	protected boolean implVerify(byte[] sigBytes, byte[] digestValue) {

		return nativeVerify(this.nativeAlgorithmId, sigBytes, sigBytes.length, this.publicKey.getNativeId(),
				digestValue, digestValue.length);
	}

	@Override
	protected byte[] implSign(byte[] digestValue) {

		int maxSize = Integer.getInteger(MAX_SIGNATURE_SIZE_PROPERTY, DEFAULT_MAX_SIGNATURE_SIZE).intValue();
		byte[] sigBytes = new byte[maxSize];

		int sigLen = nativeSign(this.nativeAlgorithmId, sigBytes, sigBytes.length, this.privateKey.getNativeId(),
				digestValue, digestValue.length);

		byte[] signature = new byte[sigLen];
		System.arraycopy(sigBytes, 0, signature, 0, sigLen);
		return signature;
	}

	@Override
	protected void setKeys(PublicKey publicKey, PrivateKey privateKey) throws InvalidKeyException {
		try {
			this.publicKey = (NativePublicKey) publicKey;
			this.privateKey = (NativePrivateKey) privateKey;
		} catch (ClassCastException e) {
			throw new InvalidKeyException(e);
		}
	}

	/**
	 * Gets for the given algorithm the signature algorithm description.
	 * <p>
	 *
	 * @param algorithm
	 *            Null terminated string that describes the algorithm
	 * @param digestAlgorithmName
	 *            (output param) Null terminated string that describes the digest algorithm.
	 * @param digestAlgorithmNameLength
	 *            (output param) Length of digest_algorithm.
	 * @return algorithm ID on success or -1 on error.
	 */
	public static native int nativeGetAlgorithmDescription(byte[] algorithm, byte[] digestAlgorithmName,
			int digestAlgorithmNameLength);

	/**
	 * Verifies the given signature.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 * @param signature
	 *            the signature to be verified
	 * @param signatureLength
	 *            the length of the signature
	 * @param nativeId
	 *            the public key native id
	 * @param digest
	 *            the digest of the data to be checked against the signature
	 * @param digestLength
	 *            the length of the digest
	 * @return true if the signature was verified, false if not
	 * @throws NativeException
	 */
	private static native boolean nativeVerify(int nativeAlgorithmId, byte[] signature, int signatureLength,
			int nativeId, byte[] digest, int digestLength) throws NativeException;

	/**
	 * Signs the given data.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 * @param signature
	 *            the array that will contain the resulting signature
	 * @param signatureLength
	 *            the length of the signature
	 * @param nativeId
	 *            the private key native id
	 * @param digest
	 *            the digest of the data to be signed
	 * @param digestLength
	 *            the length of the digest
	 * @throws NativeException
	 */
	private static native int nativeSign(int nativeAlgorithmId, byte[] signature, int signatureLength, int nativeId,
			byte[] digest, int digestLength) throws NativeException;

	/**
	 * Get algorithm OID "Object identifier".
	 *
	 * @param algorithm
	 *            algorithm name, null terminated `\0`
	 * @param oid
	 *            byte array containing a string representation of the Object identifier
	 * @param oidSize
	 *            the size of oid array
	 * @throws NativeException
	 *             on error
	 */
	public static native void getAlgOID(byte[] algorithm, byte[] oid, int oidSize) throws NativeException;

}
