/*
 * Copyright 2018-2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.microej.security.util;

import java.util.List;

/**
 *
 */
public class Arrays {

	/**
	 * Returns <tt>true</tt> if the two specified arrays of bytes are <i>equal</i> to one another. Two arrays are
	 * considered equal if both arrays contain the same number of elements, and all corresponding pairs of elements in
	 * the two arrays are equal. In other words, two arrays are equal if they contain the same elements in the same
	 * order. Also, two array references are considered equal if both are <tt>null</tt>.
	 * <p>
	 *
	 * @param a
	 *            one array to be tested for equality
	 * @param a2
	 *            the other array to be tested for equality
	 * @return <tt>true</tt> if the two arrays are equal
	 */
	public static boolean equals(byte[] a, byte[] a2) {
		if (a == a2) {
			return true;
		}
		if (a == null || a2 == null) {
			return false;
		}

		int length = a.length;
		if (a2.length != length) {
			return false;
		}

		for (int i = 0; i < length; i++) {
			if (a[i] != a2[i]) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Returns a hash code based on the contents of the specified array. For any two <tt>byte</tt> arrays <tt>a</tt> and
	 * <tt>b</tt> such that <tt>Arrays.equals(a, b)</tt>, it is also the case that
	 * <tt>Arrays.hashCode(a) == Arrays.hashCode(b)</tt>.
	 *
	 * <p>
	 * The value returned by this method is the same value that would be obtained by invoking the {@link List#hashCode()
	 * <tt>hashCode</tt>} method on a {@link List} containing a sequence of {@link Byte} instances representing the
	 * elements of <tt>a</tt> in the same order. If <tt>a</tt> is <tt>null</tt>, this method returns 0.
	 *
	 * @param a
	 *            the array whose hash value to compute
	 * @return a content-based hash code for <tt>a</tt>
	 * @since 1.5
	 */
	public static int hashCode(byte a[]) {
		if (a == null) {
			return 0;
		}

		int result = 1;
		for (byte element : a) {
			result = 31 * result + element;
		}

		return result;
	}

	/**
	 * Assigns the specified byte value to each element of the specified array of bytes.
	 *
	 * @param a
	 *            the array to be filled
	 * @param val
	 *            the value to be stored in all elements of the array
	 */
	public static void fill(byte[] a, byte val) {
		for (int i = 0, len = a.length; i < len; i++) {
			a[i] = val;
		}
	}

	/**
	 * Assigns the specified byte value to each element of the specified range of the specified array of bytes.
	 *
	 * @param a
	 *            the array to be filled
	 * @param fromIndex
	 *            start index in the array
	 * @param toIndex
	 *            end index in the array
	 * @param val
	 *            the value to be stored in all elements of the array
	 */
	public static void fill(byte[] a, int fromIndex, int toIndex, byte val) {
		for (int i = fromIndex; i < toIndex; i++) {
			a[i] = val;
		}
	}

	/**
	 * Assigns the specified char value to each element of the specified array of chars.
	 *
	 * @param a
	 *            the array to be filled
	 * @param val
	 *            the value to be stored in all elements of the array
	 */
	public static void fill(char[] a, char val) {
		for (int i = 0, len = a.length; i < len; i++) {
			a[i] = val;
		}
	}
}
