/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2021-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.microej.support.security.keypairgen;

import java.security.InvalidAlgorithmParameterException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.RSAKeyGenParameterSpec;

import com.is2t.support.security.NativePrivateKey;
import com.is2t.support.security.NativePublicKey;

import ej.sni.NativeException;
import ej.sni.NativeResource;
import ej.sni.SNI;

/**
 * NativeKeyPairGenerator Service Provider
 */
public class NativeKeyPairGeneratorSpi extends KeyPairGenerator {

	private final int nativeAlgorithmId;

	private int keySize;
	private int publicExponent;
	private String stdName;

	/**
	 * @param nativeAlgorithmId
	 * @param algorithm
	 */
	public NativeKeyPairGeneratorSpi(int nativeAlgorithmId, String algorithm) {
		super(algorithm);
		this.nativeAlgorithmId = nativeAlgorithmId;
		// default values
		// can be overridden by calling initialize method
		this.keySize = 1024;
		this.publicExponent = RSAKeyGenParameterSpec.F4;
		this.stdName = "secp521r1";
	}

	@Override
	public KeyPair generateKeyPair() {

		byte[] stdNameBytes = SNI.toCString(this.stdName);
		final int nativeId = nativeGenerateKeyPair(this.nativeAlgorithmId, this.keySize, this.publicExponent,
				stdNameBytes);

		final NativePrivateKey privateKey = new NativePrivateKey(nativeId, getAlgorithm());
		final NativePublicKey publicKey = new NativePublicKey(nativeId, getAlgorithm());

		// Only close this native resource when the private key is GC
		// The private key has a different cycle from the KeyPair object which is a simple Java holder of it
		// The public key in this context hold the same native resource that will be freed when the PrivateKey is GC
		NativeResource.closeOnGC(nativeId, nativeGetCloseId(this.nativeAlgorithmId), privateKey);
		return new KeyPair(publicKey, privateKey);
	}

	@Override
	public void initialize(int keysize, SecureRandom random) {
		this.keySize = keysize;
	}

	@Override
	public void initialize(AlgorithmParameterSpec params, SecureRandom random)
			throws InvalidAlgorithmParameterException {
		if (params instanceof RSAKeyGenParameterSpec) {
			RSAKeyGenParameterSpec p = (RSAKeyGenParameterSpec) params;
			this.keySize = p.getKeysize();
			this.publicExponent = p.getPublicExponent();

		} else if (params instanceof ECGenParameterSpec) {
			this.stdName = ((ECGenParameterSpec) params).getName();

		} else {
			throw new InvalidAlgorithmParameterException();
		}
	}

	/**
	 * Get the supported algorithm native ID.
	 *
	 * @param algorithm
	 *            Null terminated string that describes the algorithm
	 * @return algorithm ID on success or -1 on error.
	 */
	public static native int nativeGetAlgorithm(byte[] algorithm);

	/**
	 * Generate the key pair in the specific algorithm represented by algorithmId
	 *
	 * @param algorithmId
	 *            the algorithm id to use for the key generation
	 * @param rsaKeySize
	 *            the key size to be generated
	 * @param rsaPublicExponent
	 *            the public exponent for RSA keys
	 * @param ecCurveName
	 *            in order to generate the corresponding (precomputed) elliptic curve domain parameters. curve name like
	 *            secp521r1. a byte representation of the name as a NULL terminated String (ends with '\0').
	 *
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGenerateKeyPair(int algorithmId, int rsaKeySize, int rsaPublicExponent,
			byte[] ecCurveName);

	/**
	 * Gets the id of the native close function to be used as a callback in SNI close resource hookF.
	 *
	 * @param algorithmId
	 *            the algorithm id to use for the key generation
	 *
	 * @return the id of the static native close function
	 *
	 */
	private static native int nativeGetCloseId(int algorithmId);

}
