/*
 * Copyright 2023-2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */
package com.microej.support.security.secretkey;

import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;

import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactorySpi;
import javax.crypto.spec.PBEKeySpec;

import ej.sni.NativeException;
import ej.sni.NativeResource;

/**
 *
 */
public class NativeSecretKeyFactorySpi extends SecretKeyFactorySpi {
	private final String algorithm;
	private final int nativeAlgorithmId;

	/**
	 * SPI constructor.
	 *
	 * @param algorithmId
	 *            the matching native Id of the algorithm.
	 * @param algorithm
	 *            the standard string name of the algorithm.
	 */
	public NativeSecretKeyFactorySpi(int algorithmId, String algorithm) {
		this.nativeAlgorithmId = algorithmId;
		this.algorithm = algorithm;
	}

	@Override
	public SecretKey engineGenerateSecret(KeySpec keySpec) throws InvalidKeySpecException {
		PBEKeySpec encodedKeySpec = (PBEKeySpec) keySpec;
		char[] password = encodedKeySpec.getPassword();
		if (password == null) {
			// Should allow an empty password.
			password = new char[0];
		}

		byte[] passwordByte = new byte[password.length];
		for (int i = 0; i < password.length; i++) {
			passwordByte[i] = ((byte) password[i]);
		}
		byte[] salt = encodedKeySpec.getSalt();
		int iterations = encodedKeySpec.getIterationCount();
		int keyLength = encodedKeySpec.getKeyLength();

		int nativeId = nativeGetKeyData(this.nativeAlgorithmId, passwordByte, passwordByte.length, salt, salt.length,
				iterations, keyLength);
		SecretKey secretKey = new NativeSecretKey(nativeId, this.algorithm);
		NativeResource.closeOnGC(nativeId, nativeGetCloseId(this.nativeAlgorithmId), secretKey);
		return secretKey;
	}

	/**
	 * Get the supported algorithm native ID.
	 *
	 * @param algorithm
	 *            Null terminated string that describes the algorithm
	 * @return algorithm ID on success or -1 on error.
	 */
	public static native int nativeGetAlgorithm(byte[] algorithm);

	/**
	 * Generate a public key from the encoded key format.
	 *
	 * @param password
	 *            the password to encode, a null terminated '\0' byte array representation of the format in String
	 * @param passwordLength
	 *            the length of password
	 * @param algorithmId
	 *            the algorithm ID
	 * @param salt
	 *            salty
	 * @param saltLength
	 *            salty
	 * @param iterations
	 *            number of iterations
	 * @param keyLength
	 *            the length of encodedKey to generate (in bits)
	 * @return the pointer of the C structure holding the key data
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetKeyData(int algorithmId, byte[] password, int passwordLength, byte[] salt,
			int saltLength, int iterations, int keyLength);

	/**
	 * Gets the id of the native close function.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetCloseId(int algorithmId);

}
