/*
 * Copyright (c) 1996, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2015-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package java.net;

import java.io.IOException;

import ej.lang.Resource;

/**
 * Abstract datagram and multicast socket implementation base class.
 * 
 * @author Pavani Diwanji
 * @since JDK1.1
 */

public abstract class DatagramSocketImpl implements SocketOptions, Resource {

	/**
	 * The local port number.
	 */
	protected int localPort;

	/**
	 * Creates a datagram socket.
	 * 
	 * @exception SocketException
	 *                if there is an error in the underlying protocol, such as a TCP error.
	 */
	protected abstract void create() throws SocketException;

	/**
	 * Binds a datagram socket to a local port and address.
	 * 
	 * @param lport
	 *            the local port
	 * @param laddr
	 *            the local address
	 * @exception SocketException
	 *                if there is an error in the underlying protocol, such as a TCP error.
	 */
	protected abstract void bind(int lport, InetAddress laddr) throws SocketException;

	/**
	 * Sends a datagram packet. The packet contains the data and the destination address to send the packet to.
	 * 
	 * @param p
	 *            the packet to be sent.
	 * @exception IOException
	 *                if an I/O exception occurs while sending the datagram packet.
	 * @exception PortUnreachableException
	 *                may be thrown if the socket is connected to a currently unreachable destination. Note, there is no
	 *                guarantee that the exception will be thrown.
	 */
	protected abstract void send(DatagramPacket p) throws IOException;

	/**
	 * Connects a datagram socket to a remote destination. This associates the remote address with the local socket so
	 * that datagrams may only be sent to this destination and received from this destination. This may be overridden to
	 * call a native system connect.
	 *
	 * <p>
	 * If the remote destination to which the socket is connected does not exist, or is otherwise unreachable, and if an
	 * ICMP destination unreachable packet has been received for that address, then a subsequent call to send or receive
	 * may throw a PortUnreachableException. Note, there is no guarantee that the exception will be thrown.
	 * 
	 * @param address
	 *            the remote InetAddress to connect to
	 * @param port
	 *            the remote port number
	 * @exception SocketException
	 *                may be thrown if the socket cannot be connected to the remote destination
	 * @since 1.4
	 */
	protected void connect(InetAddress address, int port) throws SocketException {
	}

	/**
	 * Disconnects a datagram socket from its remote destination.
	 * 
	 * @since 1.4
	 */
	protected void disconnect() {
	}

	protected abstract void receive(DatagramPacket p) throws IOException;

	protected abstract void setTimeToLive(int ttl) throws IOException;

	protected abstract int getTimeToLive() throws IOException;

	protected abstract void join(InetAddress inetaddr) throws IOException;

	protected abstract void leave(InetAddress inetaddr) throws IOException;

	protected abstract void joinGroup(SocketAddress mcastaddr, NetworkInterface netIf) throws IOException;

	protected abstract void leaveGroup(SocketAddress mcastaddr, NetworkInterface netIf) throws IOException;

	/**
	 * Close the socket.
	 */
	protected abstract void close();

	/**
	 * Gets the local port.
	 * 
	 * @return an {@code int} representing the local port value
	 */
	protected int getLocalPort() {
		return localPort;
	}

	@Override
	public Object getSource() {
		return this;
	}

	@Override
	public void reclaim() {
		close();
	}
}
